import logging
import warnings
import numpy as np
from pyspedas.tplot_tools import get, store, options, get_data, store_data

# use nanmean from bottleneck if it's installed, otherwise use the numpy one
# bottleneck nanmean is ~2.5x faster
try:
    import bottleneck as bn
    nanmean = bn.nanmean
except ImportError:
    nanmean = np.nanmean

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)


def mms_feeps_omni(eyes,
                   probe='1',
                   datatype='electron',
                   data_units='intensity',
                   data_rate='srvy',
                   level='l2',
                   suffix='',
                   quality_flag=3,
                   get_err=False,
                   ):
    """
    This function will calculate the omnidirectional FEEPS spectrograms, and is automatically called from mms_load_feeps
    
    Parameters
    ------------
        eyes: dict
            Hash table containing the active sensor eyes

        probe: str
            probe #, e.g., '4' for MMS4

        datatype: str
            'electron' or 'ion'

        data_units: str
            'intensity'

        data_rate: str
            instrument data rate, e.g., 'srvy' or 'brst'

        level: str
            data level

        suffix: str
            suffix of the loaded data

        quality_flag: int
            Value can range from 2-4. Determines which eyes to use to calculate omnidirectional flux at a given time
            based on how many sectors were masked for sunlight. For BRST or RAW::

                Green (Quality Indicator = 0): No contaminated or masked spin sectors (best data quality)
                Red (Quality Indicator = 3): Contaminated spin sector is associated with BURST and RAW data
                    (this data is not recommended for scientific use)
                Grey (Quality Indicator = 4): FEEPS calibration data (this data is generated by the instrument itself
                    and is not recommended for scientific use)

            For data in SURVEY mode, there are 5 possible quality indicators::

                Green (Quality Indicator = 0): No contaminated or masked spin sectors (best data quality)
                Yellow (Quality Indicator = 1): Among the spin sectors used to create a SURVEY aggregate
                     spin sector, onboard masking is successfully applied to a minimum of 1 and a maximum of 7
                     spin sectors and there is no contamination in any of the individual spin sectors. Note that
                     since this indicator applies only when onboard masking is applied, some sectors are not
                     observed.
                Orange (Quality Indicator = 2): There is contamination of between 0 and 50% of the notmasked sectors. Warning: this data contains contamination.
                Red (Quality Indicator = 3): Same as orange except that there is contamination of 50% or
                     more of the not-masked sectors. IN ADDITION, this indicator applies to the case where
                     onboard masking is applied to all 8 spin sectors in the aggregate sector.
                Grey (Quality Indicator = 4): FEEPS calibration data (this data is generated by the instrument
                     itself and is not recommended for scientific use)

        get_err: bool
            Gets the error in the omnidirectional l2 intensity.  Default: False. User MUST run mms_load_feeps for the
            l1b counts data FIRST to use this function. (Note: the FEEPS l1b data needed for this switch is restricted
            to MMS team members at present.)

    Returns
    ------------
        List of tplot variables created.
    """
    out_vars = []
    units_label = ''
    if data_units == 'intensity':
        units_label = '1/(cm^2-sr-s-keV)'
    elif data_units == 'counts':
        units_label = '[counts/s]'

    prefix = 'mms'+probe+'_epd_feeps_'
    if datatype == 'electron':
        energies = np.array([33.2, 51.90, 70.6, 89.4, 107.1, 125.2, 146.5, 171.3,
                    200.2, 234.0, 273.4, 319.4, 373.2, 436.0, 509.2])
    else:
        energies = np.array([57.9, 76.8, 95.4, 114.1, 133.0, 153.7, 177.6,
                    205.1, 236.7, 273.2, 315.4, 363.8, 419.7, 484.2,  558.6])

    en_label=energies

    # set unique energy bins per spacecraft; from DLT on 31 Jan 2017
    eEcorr = [14.0, -1.0, -3.0, -3.0]
    iEcorr = [0.0, 0.0, 0.0, 0.0]
    eGfact = [1.0, 1.0, 1.0, 1.0]
    iGfact = [0.84, 1.0, 1.0, 1.0]

    if probe == '1' and datatype == 'electron':
        energies = energies + eEcorr[0]
    if probe == '2' and datatype == 'electron':
        energies = energies + eEcorr[1]
    if probe == '3' and datatype == 'electron':
        energies = energies + eEcorr[2]
    if probe == '4' and datatype == 'electron':
        energies = energies + eEcorr[3]

    if probe == '1' and datatype == 'ion':
        energies = energies + iEcorr[0]
    if probe == '2' and datatype == 'ion':
        energies = energies + iEcorr[1]
    if probe == '3' and datatype == 'ion':
        energies = energies + iEcorr[2]
    if probe == '4' and datatype == 'ion':
        energies = energies + iEcorr[3]

    # percent error around energy bin center to accept data for averaging; 
    # anything outside of energies[i] +/- en_chk*energies[i] will be changed 
    # to NAN and not averaged   
    en_chk = 0.10

    top_sensors = eyes['top']
    bot_sensors = eyes['bottom']

    tmpdata = get(prefix+data_rate+'_'+level+'_'+datatype+'_top_'+data_units+'_sensorid_'+str(top_sensors[0])+'_clean_sun_removed'+suffix)

    if get_err:
        counts_arr = np.full((len(tmpdata.times),len(tmpdata.v),len(top_sensors)+len(bot_sensors)), np.nan)

    if tmpdata is not None:
        if level != 'sitl':
            dalleyes = np.empty((len(tmpdata[0]), len(tmpdata[2]), len(top_sensors)+len(bot_sensors)))
            dalleyes[:] = np.nan

            for idx, sensor in enumerate(top_sensors):
                var_name = prefix+data_rate+'_'+level+'_'+datatype+'_top_'+data_units+'_sensorid_'+str(sensor)+'_clean_sun_removed'+suffix
                data = get(var_name)
                dalleyes[:, :, idx] = data[1]
                try:
                    iE = np.where(np.abs(energies-data[2]) > en_chk*energies)
                    if iE[0].size != 0:
                        dalleyes[:, iE[0], idx] = np.nan
                except Warning:
                    logging.warning('NaN in energy table encountered; sensor T' + str(sensor))

                if level == 'l2':
                    var_name2 = prefix+data_rate+'_'+level+'_'+datatype+'_top_quality_indicator_sensorid_'+str(sensor)+suffix
                    dat=get_data(var_name2)
                    bad_quality=np.where(dat.y >= quality_flag)
                    if bad_quality[0].size != 0:
                        dalleyes[bad_quality[0],:,idx] = np.nan

                    if get_err:
                        var_name3=prefix+data_rate+'_l1b_'+datatype+'_top_counts_sensorid_'+str(sensor)+'_clean_sun_removed'+suffix
                        d3=get_data(var_name3)
                        counts_arr[:,:,idx] = d3.y
                        if iE[0].size != 0:
                            counts_arr[:, iE[0], idx] = np.nan
                        if bad_quality[0].size != 0:
                            counts_arr[bad_quality[0], :, idx] = np.nan

            for idx, sensor in enumerate(bot_sensors):
                var_name = prefix+data_rate+'_'+level+'_'+datatype+'_bottom_'+data_units+'_sensorid_'+str(sensor)+'_clean_sun_removed'+suffix
                data = get(var_name)
                dalleyes[:, :, idx+len(top_sensors)] = data[1]
                try:
                    iE = np.where(np.abs(energies-data[2]) > en_chk*energies)
                    if iE[0].size != 0:
                        dalleyes[:, iE[0], idx+len(top_sensors)] = np.nan
                except Warning:
                    logging.warning('NaN in energy table encountered; sensor B' + str(sensor))

                if level == 'l2':
                    var_name2 = prefix+data_rate+'_'+level+'_'+datatype+'_bottom_quality_indicator_sensorid_'+str(sensor)+suffix
                    dat=get_data(var_name2)
                    bad_quality=np.where(dat.y >= quality_flag)
                    if bad_quality[0].size != 0:
                        dalleyes[bad_quality[0],:,idx+len(top_sensors)] = np.nan

                    if get_err:
                        var_name3=prefix+data_rate+'_l1b_'+datatype+'_bottom_counts_sensorid_'+str(sensor)+'_clean_sun_removed'+suffix
                        d3=get_data(var_name3)
                        counts_arr[:,:,idx+len(top_sensors)] = d3.y
                        if iE[0].size != 0:
                            counts_arr[:, iE[0], idx+len(top_sensors)] = np.nan
                        if bad_quality[0].size != 0:
                            counts_arr[bad_quality[0], :, idx+len(top_sensors)] = np.nan

        else:  # sitl data
            dalleyes = np.empty((len(tmpdata[0]), len(tmpdata[2]), len(top_sensors)))
            dalleyes[:] = np.nan

            # Only top sensors in SITL product
            for idx, sensor in enumerate(top_sensors):
                var_name = prefix+data_rate+'_'+level+'_'+datatype+'_top_'+data_units+'_sensorid_'+str(sensor)+'_clean_sun_removed'+suffix
                data = get(var_name)
                dalleyes[:, :, idx] = data[1]
                iE = np.where(np.abs(energies-data[2]) > en_chk*energies)
                if iE[0].size != 0:
                    dalleyes[:, iE[0], idx] = np.nan

                if level == 'l2':
                    var_name2 = prefix+data_rate+'_'+level+'_'+datatype+'_top_quality_indicator_sensorid_'+str(sensor)+suffix
                    dat=get_data(var_name2)
                    bad_quality=np.where(dat.y >= quality_flag)
                    if bad_quality[0].size != 0:
                        dalleyes[bad_quality[0],:,idx] = np.nan

                    if get_err:
                        var_name3=prefix+data_rate+'_l1b_'+datatype+'_top_counts_sensorid_'+str(sensor)+'_clean_sun_removed'+suffix
                        d3=get_data(var_name3)
                        counts_arr[:,:,idx] = d3.y
                        if iE[0].size != 0:
                            counts_arr[:, iE[0], idx] = np.nan
                        if bad_quality[0].size != 0:
                            counts_arr[bad_quality[0], :, idx] = np.nan

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            flux_omni = nanmean(dalleyes, axis=2)

        if probe == '1' and datatype == 'electron':
            flux_omni = flux_omni*eGfact[0]
        if probe == '2' and datatype == 'electron':
            flux_omni = flux_omni*eGfact[1]
        if probe == '3' and datatype == 'electron':
            flux_omni = flux_omni*eGfact[2]
        if probe == '4' and datatype == 'electron':
            flux_omni = flux_omni*eGfact[3]

        if probe == '1' and datatype == 'ion':
            flux_omni = flux_omni*iGfact[0]
        if probe == '2' and datatype == 'ion':
            flux_omni = flux_omni*iGfact[1]
        if probe == '3' and datatype == 'ion':
            flux_omni = flux_omni*iGfact[2]
        if probe == '4' and datatype == 'ion':
            flux_omni = flux_omni*iGfact[3]

        store('mms'+probe+'_epd_feeps_'+data_rate+'_'+level+'_'+datatype+'_'+data_units+'_omni'+suffix, data={'x': tmpdata[0], 'y': flux_omni, 'v': energies})
        options('mms'+probe+'_epd_feeps_'+data_rate+'_'+level+'_'+datatype+'_'+data_units+'_omni'+suffix, 'spec', True)
        options('mms'+probe+'_epd_feeps_'+data_rate+'_'+level+'_'+datatype+'_'+data_units+'_omni'+suffix, 'ylog', True)
        options('mms'+probe+'_epd_feeps_'+data_rate+'_'+level+'_'+datatype+'_'+data_units+'_omni'+suffix, 'zlog', True)
        options('mms'+probe+'_epd_feeps_'+data_rate+'_'+level+'_'+datatype+'_'+data_units+'_omni'+suffix, 'ztitle', units_label)
        options('mms'+probe+'_epd_feeps_'+data_rate+'_'+level+'_'+datatype+'_'+data_units+'_omni'+suffix, 'ytitle', 'MMS' + str(probe) + ' ' + datatype)
        options('mms'+probe+'_epd_feeps_'+data_rate+'_'+level+'_'+datatype+'_'+data_units+'_omni'+suffix, 'ysubtitle', '[keV]')
        out_vars.append('mms'+probe+'_epd_feeps_'+data_rate+'_'+level+'_'+datatype+'_'+data_units+'_omni'+suffix)

        if get_err:
            uncertainty_arr = np.sqrt(counts_arr) / counts_arr # This is the percent uncertainty for each[time, energy, eye]
            uncertainty_arr_sq = uncertainty_arr*uncertainty_arr
            sqrt_uncertainty_arr_sum_sq = np.sqrt(np.nansum(uncertainty_arr_sq, 2))
            finite_divisor = np.isfinite(uncertainty_arr_sq) # Find out how many eyes were used. For example, if only 2 eyes had counts, you want to average over 2 eyes, not over 18 eyes.
            divisor = np.sum(finite_divisor, 2)
            percent_uncertainty = sqrt_uncertainty_arr_sum_sq / divisor

            # percent_uncertainty = reform(average(uncertainty_arr_sq, 3, / NAN))
            newname2 = 'mms' + probe + '_epd_feeps_' + data_rate + '_' + level + '_' + datatype + '_' + data_units + '_omni_percent_uncertainty' + suffix
            # not units of percent. So the value here of  0.1 is 10 %.calculated above as sqrt(N) / N.
            store_data(newname2, data = {'x': tmpdata.times, 'y': percent_uncertainty,'v': en_label})
            options(newname2, 'spec',1)
            options(newname2, 'ylog',1)
            options(newname2, 'zlog',1)
            options(newname2, 'ztitle','[Percent uncertainty/100]')
            options(newname2, 'ysubtitle','[keV]')
            out_vars.append(newname2)

    return out_vars


# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/Pytplot

import pyspedas
from pyspedas.tplot_tools import store_data, tnames
import numpy as np
import copy
import logging

def clip(tvar,ymin,ymax,newname=None):
    """
    Change out-of-bounds data to NaN.

    Parameters
    ----------
        tvar : str or list[str]
            tplot variable names to clip (wildcards accepted)
        ymin : int/float
            Minimum value to keep (inclusive)
        ymax : int/float
            Maximum value to keep (inclusive)
        newname : str
            Name of new tvar for clipped data storage.  If not specified, tvar will be replaced
            THIS is not an option for multiple variable input, for multiple or pseudo variables, the data is overwritten.

    Returns
    -------
        None

    Examples
    --------

        >>> Make any values below 2 and above 6 equal to NaN.
        >>> pyspedas.store_data('d', data={'x':[2,5,8,11,14,17,21], 'y':[[1,1],[2,2],[100,100],[4,4],[5,5],[6,6],[7,7]]})
        >>> pyspedas.clip('d',2,6,'e')

    """

    #check for globbed or array input, and call recursively
    tn = tnames(tvar)
    if len(tn) == 0:
        return
    elif len(tn) > 1:
        for j in range(len(tn)):
            clip(tn[j],ymin,ymax)
        return


    a = copy.deepcopy(pyspedas.tplot_tools.data_quants[tvar].where(pyspedas.tplot_tools.data_quants[tvar] >= ymin))
    a = copy.deepcopy(a.where(a <= ymax))

    if newname is None:
        a.name = tvar
        pyspedas.tplot_tools.data_quants[tvar] = a
    else:
        if 'spec_bins' in a.coords:
            store_data(newname, data={'x': a.coords['time'], 'y': a.values, 'v': a.coords['spec_bins']})
            pyspedas.tplot_tools.data_quants[newname].attrs = copy.deepcopy(pyspedas.tplot_tools.data_quants[tvar].attrs)
        else:
            store_data(newname, data={'x': a.coords['time'], 'y': a.values})
            pyspedas.tplot_tools.data_quants[newname].attrs = copy.deepcopy(pyspedas.tplot_tools.data_quants[tvar].attrs)

    return



import logging
from pyspedas.tplot_tools import get_data, store_data, options
from pyspedas.tplot_tools import tnames


def rbsp_load_rbspice_read(level='l3', probe='a', datatype='TOFxEH'):
    """
    Works on previously loaded RBSPICE tplot variables: adds energy channel energy values to primary data variable, separates non-H species variables,
    creates variables for individual telescopes, and sets appropriate tplot options
    
    Parameters
    ----------
    level : str, default='l3'
        data level 'l1','l2','l3'

    probe : str or list of str, default='a'
        Spacecraft probe name: 'a' or 'b'

    datatype : str, default='TOFxEH'
        RBSPICE data type: 'EBR','ESRHELT','ESRLEHT','IBR','ISBR','ISRHELT','TOFxEH' (default),'TOFxEIon','TOFxEnonH','TOFxPHHHELT','TOFxPHHLEHT'
        but change for different data levels.

    Returns
    -------
    None

    Examples
    --------
    This function is called within pyspedas.projects.rbsp.rbspice
    """
    if level != 'l1':
        units_label = '1/(cm^2-sr-s-keV)'
        convert_factor = 1000.              # to convert flux from 1/MeV to 1/keV
    else:
        units_label = 'counts/s'
        convert_factor = 1.                 # do not need to convert counts/s           

    prefix = 'rbsp'+probe+'_rbspice_'+level+'_'+datatype+'_'
    # find the flux/cps data name(s)
    data_var = tnames(prefix + 'F*DU')
    energy_var = tnames(prefix + 'F*DU_Energy')

    logging.info('Correcting RBSPICE energy tables...')
    for i in range(len(data_var)):
        en_data = get_data(energy_var[i])
        temp_energy = en_data.transpose()
        temp = get_data(data_var[i])
        data = temp.y.transpose([0, 2, 1])
        species_str = data_var[i][-4:-2]
        if species_str == 'FP':
            species='proton'
            yticks = 1
            if datatype != 'TOFxPHHHELT':
                new_energy = temp_energy[:,0] * 1000.           # convert energy from MeV to keV
                new_flux = data / convert_factor                # convert flux from 1/MeV to 1/keV
                zrange = [5.,1.e5]
            else:
                new_energy = temp_energy[11:-1,0] * 1000.       # convert energy from MeV to keV
                new_flux = data[:,11:-1,:] / convert_factor       # convert energy from MeV to keV
                zrange = [2.e2,1.e6]
        elif species_str == 'He':
            species='helium'
            yticks = 1
            new_energy = temp_energy[0:10,0] * 1000.          # convert energy from MeV to keV
            new_flux = data[:,0:10,:] / convert_factor        # convert flux from 1/MeV to 1/keV
            zrange = [1.,1.e3]
        elif species_str == 'FO':
            species='oxygen'
            yticks = 2
            if datatype != 'TOFxPHHHELT':
                new_energy = temp_energy[11:18,0] * 1000.       # convert energy from MeV to keV
                new_flux = data[:,11:18,:] / convert_factor     # convert flux from 1/MeV to 1/keV
                zrange = [1.,1.e2]
            else:
                new_energy = temp_energy[0:10,0] * 1000.        # convert energy from MeV to keV
                new_flux = data[:,0:10,:] / convert_factor      # convert flux from 1/MeV to 1/keV
                zrange = [1e1,1.e4]
        new_name = prefix+species
        # note: can't save the energy table here, due to the shape of new_flux
        # so we'll have to grab the energy table from the individual telescope
        # variables
        store_data(new_name, data={'x':temp.times, 'y':new_flux})
        options(new_name, 'ylog', True)
        options(new_name, 'zlog', True)
        options(new_name, 'zrange', zrange)
        options(new_name, 'ytitle', 'rbsp'+probe+'_rbspice_'+species)
        options(new_name, 'ysubtitle', 'Energy [keV]')
        options(new_name, 'ztitle', units_label)
        for j in range(6):
            store_data(new_name+'_T'+str(j), data={'x':temp.times, 'y':new_flux[:,:,j], 'v':new_energy})
            options(new_name+'_T'+str(j), 'spec', True)
            options(new_name+'_T'+str(j), 'ylog', True)
            options(new_name+'_T'+str(j), 'zlog', True)
            options(new_name+'_T'+str(j), 'zrange', zrange)
            options(new_name+'_T'+str(j), 'ytitle', 'rbsp'+probe+'_rbspice_'+species+'_T'+str(j))
            options(new_name+'_T'+str(j), 'ysubtitle', '[keV]')
            options(new_name+'_T'+str(j), 'ztitle', units_label)




from pyspedas.tplot_tools import timebar
import pyspedas
import logging


def databar(varname, y=0.0, delete=False, color="black", thick=1, dash=False):
    """
    This function will add a horizontal bar at the given value to a plot of a tplot variable.

    Parameters
    ----------
    y : float or list
        The Y-axis value  If a list of numbers are supplied, multiple bars will be created.
    varname : str or list
        The variable(s) to add the vertical bar to. (wildcards accepted)
    delete : bool, optional
        If set to True, at least one varname must be supplied. The timebar at value "y" for variable "varname"
        will be removed.
    color : str
        The color of the bar.
    thick : int
        The thickness of the bar.
    dash : bool
        If set to True, the bar is dashed rather than solid.

    Returns
    -------
    None

    Examples
    --------
    >>> # Place a green time bar at 2017-07-17 00:00:00
    >>> import pyspedas
    >>> pyspedas.databar('*', 0, color='green')

    >>> # Place a dashed data bar at 5500 on the y axis
    >>> pyspedas.databar('*', 5500, dash=True)

    >>> # Place 3 magenta time bars of thickness 5
    >>> # at -100, 0, +100
    >>> # for variable 'sgx' plot
    >>> pyspedas.databar('sgx', [-100, 0, 100],color='m',thick=5)
    """

    # wildcard expansion
    varname = pyspedas.tnames(varname)

    if len(varname) == 0:
        logging.warning("No valid tplot variables specified")
        return

    return timebar(y, varname, databar=True, delete=delete, color=color, thick=thick, dash=dash)



from ftplib import FTP
import os
import time
import logging


def download_ftp(
    ftp_server,
    remote_path,
    remote_file,
    local_path,
    local_file=None,
    username="anonymous",
    password="anonymous@",
    force_download=False,
):
    """
    Download a file from an FTP server.

    Parameters
    ----------
    ftp_server : str
        FTP server name or IP address.
    remote_path : str
        Path on the FTP server where the file is located.
    remote_file : str
        Name of the file to download.
    local_path : str
        Local directory to save the file.
    local_file : str, optional
        Name of the file to save locally. If not provided, the name of the remote file is used.
    username : str, optional
        Username for the FTP server. Default is 'anonymous'.
    password : str, optional
        Password for the FTP server. Default is 'anonymous@'.
    force_download : bool, optional
        Force the download even if the remote file is not newer than the local file. Default is False.

    Returns
    -------
    list
        A list containing the path of the downloaded file.

    Raises
    ------
    Exception
        If the remote file is not found on the FTP server.

    Examples
    --------
    >>> from pyspedas import download_ftp
    >>> ftp_site = "ftp.gfz-potsdam.de"
    >>> kp_dir = "/pub/home/obs/kp-ap/wdc/yearly/"
    >>> remote_file = "kp2012.wdc"
    >>> local_dir = "/tmp/"
    >>> files = download_ftp(ftp_site, kp_dir, remote_file, local_dir)
    >>> print(files)
    ['/tmp/kp2012.wdc']
    """
    return_files = []

    try:
        with FTP(ftp_server) as ftp:
            ftp.login(user=username, passwd=password)
            ftp.cwd(remote_path)  # Change to the directory containing the file

            if local_file is None:
                local_file = os.path.join(local_path, remote_file)

            # Check if remote file exists
            if remote_file not in ftp.nlst():
                msg = f"File '{remote_file}' was not found on the FTP server '{ftp_server}'"
                raise Exception(msg)

            # Get the modification time of the remote file
            response = ftp.sendcmd("MDTM " + remote_file)
            remote_mtime = time.mktime(time.strptime(response[4:], "%Y%m%d%H%M%S"))

            # Get the modification time of the local file
            local_mtime = (
                os.path.getmtime(local_file) if os.path.exists(local_file) else 0
            )

            # Get the directory name from the file path
            dir_name = os.path.dirname(local_file)

            # Create the directory if it doesn't exist
            os.makedirs(dir_name, exist_ok=True)

            # Download the file if the remote file has been changed or if force_download is True
            if force_download or remote_mtime > local_mtime:
                with open(local_file, "wb") as local_file_r:
                    ftp.retrbinary("RETR " + remote_file, local_file_r.write)
                logging.warning(
                    f"File '{remote_file}' downloaded successfully to '{local_file}'"
                )
            else:
                logging.warning(
                    f"File '{remote_file}' has not been modified since last download"
                )
        return_files.append(local_file)
    except Exception as e:
        logging.error(e)

    return return_files



import numpy as np
import logging
import pyspedas
from pyspedas.tplot_tools import data_exists, store_data, get_data, options
from scipy import signal


# First pass at the power spectrum function.  This is still missing several features of the IDL power spectrum routine, such as
# bin, nohanning, notperhertz, and tm_sensativity.  The IDL routine is located in dpwrspc.pro.

# There is also the issue of this not quite having the same units as the plot I use as my reference.
# https://agupubs.onlinelibrary.wiley.com/doi/full/10.1002/2015GL065366#grl53372-bib-0016
# Interestingly enough, the output is the same if units of seconds are used in the periodogram instead of Hertz.
# Perhaps they calculated it differently?


def pwr_spec(tvar, nbp=256, nsp=128, newname=None):
    """
    Calculates the power spectrum of a line, and adds a tplot variable for this new spectrogram

    Parameters
    ----------
        tvar : str
            Name of tvar to use
        nbp : int, optional
            The number of points to use when calculating the FFT
        nsp : int, optional
            The number of points to shift over to calculate the next FFT
        newname : str, optional
            The name of the new tplot variable created,

    Returns
    -------
        None

    Examples
    --------
        >>> import pyspedas
        >>> import math
        >>> time = [pyspedas.time_float("2020-01-01") + i for i in range(10000)]
        >>> quantity = [math.sin(i) for i in range(10000)]
        >>> pyspedas.store_data("dp", data={"x": time, "y": quantity})
        >>> pyspedas.pwr_spec("dp", newname="dp_pwrspec")
        >>> pyspedas.tplot("dp_pwrspec")
    """

    if not data_exists(tvar):
        logging.error("Input variable %s does not exist", tvar)
        return

    d =get_data(tvar)
    x, y = d[0], d[1]

    if len(y.shape) > 1:
        logging.warning(
            "Cannot create pwr_spec for variable %s, must be a single line", tvar
        )

    l = len(x)
    x_new = []
    f_new = []
    pxx_new = []
    shift_lsp = np.arange(0, l - 1, nsp)
    for i in shift_lsp:
        x_n = x[i : i + nbp]
        y_n = y[i : i + nbp]
        if len(x_n) < nbp:
            continue

        median_diff_between_points = np.median(np.diff(x_n))
        w = signal.get_window("hamming", nbp)
        f, pxx = signal.periodogram(
            y_n, fs=(1 / median_diff_between_points), window=w, detrend="linear"
        )
        f = f[1:-1]
        pxx = pxx[1:-1]
        x_new.append((x_n[-1] + x_n[0]) / 2)
        f_new.append(f)
        pxx_new.append(pxx)

    if newname is None:
        newname = tvar + "_pwrspec"

    store_data(newname, data={"x": x_new, "y": pxx_new, "v": f_new})
    options(newname, "spec", 1)
    options(newname, "zlog", 1)
    options(newname, "ylog", 1)

    return




from .load import load

# This routine was originally in psp/__init__.py.  If you need to see the history of this routine before
# it was moved to its own file, please check the history for __init__.py.

# Loading
def spi(trange=['2018-11-5', '2018-11-6'], 
        datatype='sf00_l3_mom', 
        level='l3',
        suffix='',  
        prefix='',
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False,
        username=None,
        password=None,
        last_version=False,
        force_download=False,
        version=None,
    ):
    """
    This function loads Parker Solar Probe SWEAP/SPAN-i data
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options Include:
                'sf00_l3_mom': Moments of the Proton distribution function (RTN)
                'sf0a_l3_mom': Moments of the Alpha distribution function (RTN)
                'sf00_l3_mom_inst': Moments of the Proton distribution function (Instrument Frame)
                'sf0a_l3_mom_inst': Moments of the Alpha distribution function (Instrument Frame)

        suffix: str
            The tplot variable names will be given this suffix. By default,
            no prefix is added.
            Default: ''

        prefix: str
            The tplot variable names will be given this prefix.  By default,
            no prefix is added.
            Default: ''

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

        username: str
            Username to use for authentication.
            
            If passed, attempt to download data from the SWEAP Instrument Team server
            instead of the fully public server at SPDF.
            Provides access to unpublished files.

        password: str
            Password to use for authentication

        last_version: bool
            If True, only download the highest-numbered file version
        
        force_download: bool
            If True, downloads the file even if a newer version exists locally. 
            Default: False.

        version: str
            If supplied, the load routine will look for this specific CDF version, rather than a wildcard pattern
            matching any version.  Valid version strings look like "vN.M" for FIELDS sqtn_rfs_v1v1 data, or
            "vNN" for all other data types. Default: None

    Returns
    ----------
        List of tplot variables created.

    """
    if datatype in ['sf00_l3_mom','sf0a_l3_mom','sf00_l3_mom_inst','sf0a_l3_mom_inst']:
        datatype = 'spi_' + datatype
        level = 'l3'
        print("Using LEVEL=L3")

    return load(instrument='spi', trange=trange, datatype=datatype, level=level, 
        suffix=suffix, prefix=prefix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, 
        downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, 
        username=username, password=password,last_version=last_version, force_download=force_download,
        version=version
        )




import pyspedas
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import LinearSegmentedColormap
from .slice2d_getinfo import slice2d_getinfo


def slice2d_plot(
    the_slice,
    xrange=None,
    yrange=None,
    zrange=None,
    colormap=None,
    olines=8,
    contours=False,
    plotsize=10,
    title=None,
    save_png=None,
    save_jpeg=None,
    save_svg=None,
    save_pdf=None,
    save_eps=None,
    dpi=None,
    display=True,
):
    """
    Creates plots of 2D particle slices
    """
    spec_options = {}

    if zrange is None:
        zrange = the_slice.get("zrange")
        # padding, # in thm_ui_slice2d should match this
        zrange = [zrange[0] * 0.999, zrange[1] * 1.001]

    if zrange is None:
        spec_options["norm"] = mpl.colors.LogNorm()
    else:
        spec_options["norm"] = mpl.colors.LogNorm(vmin=zrange[0], vmax=zrange[1])

    style = pyspedas.tplot_tools.tplot_opt_glob.get("style")

    if style is None:
        if colormap is None:
            colormap = "spedas"
    else:
        plt.style.use(style)

    if colormap == "spedas":
        _colors = pyspedas.tplot_tools.spedas_colorbar
        spd_map = [
            (np.array([r, g, b])).astype(np.float64) / 256
            for r, g, b in zip(_colors.r, _colors.g, _colors.b)
        ]
        cmap = LinearSegmentedColormap.from_list("spedas", spd_map)
    else:
        cmap = colormap

    spec_options["cmap"] = cmap

    char_size = pyspedas.tplot_tools.tplot_opt_glob.get("charsize")
    if char_size is None:
        char_size = 12

    fig, axes = plt.subplots()
    fig.set_size_inches(plotsize, plotsize)

    if xrange is not None:
        axes.set_xlim(xrange)

    if yrange is not None:
        axes.set_ylim(yrange)

    axis_font_size = pyspedas.tplot_tools.tplot_opt_glob.get("axis_font_size")

    if axis_font_size is not None:
        axes.tick_params(axis="x", labelsize=axis_font_size)
        axes.tick_params(axis="y", labelsize=axis_font_size)

    info = slice2d_getinfo(the_slice, title=title)

    axes.set_title(info["title"], fontsize=char_size)
    axes.set_ylabel(info["ytitle"], fontsize=char_size)
    axes.set_xlabel(info["xtitle"], fontsize=char_size)

    fig.subplots_adjust(left=0.14, right=0.86, top=0.86, bottom=0.14)

    box = axes.get_position()
    pad, width = 0.02, 0.01
    cax = fig.add_axes([box.xmax + pad, box.ymin, width, box.height])

    if axis_font_size is not None:
        cax.tick_params(labelsize=axis_font_size)

    im = axes.pcolormesh(
        the_slice["xgrid"], the_slice["ygrid"], the_slice["data"].T, **spec_options
    )

    colorbar = fig.colorbar(im, cax=cax)

    colorbar.set_label(info["ztitle"], fontsize=char_size)

    # draw lines at the origin
    axes.axvline(x=0, linestyle=(0, (5, 10)), color="black")
    axes.axhline(y=0, linestyle=(0, (5, 10)), color="black")

    interp_type = the_slice.get("interpolation")

    if interp_type != "geometric" or contours:
        levels = 10.0 ** (
            np.arange(olines)
            / float(olines)
            * (np.log10(zrange[1]) - np.log10(zrange[0]))
            + np.log10(zrange[0])
        )
        contours = axes.contour(
            the_slice["xgrid"],
            the_slice["ygrid"],
            the_slice["data"].T,
            levels,
            linewidths=0.5,
        )

    if save_png is not None and save_png != "":
        if not save_png.endswith(".png"):
            save_png += ".png"
        plt.savefig(save_png, dpi=dpi)

    if save_jpeg is not None and save_jpeg != "":
        if not save_jpeg.endswith(".jpeg"):
            save_jpeg += ".jpeg"
        plt.savefig(save_jpeg, dpi=dpi)

    if save_eps is not None and save_eps != "":
        if not save_eps.endswith(".eps"):
            save_eps += ".eps"
        plt.savefig(save_eps, dpi=dpi)

    if save_svg is not None and save_svg != "":
        if not save_svg.endswith(".svg"):
            save_svg += ".svg"
        plt.savefig(save_svg, dpi=dpi)

    if save_pdf is not None and save_pdf != "":
        if not save_pdf.endswith(".pdf"):
            save_pdf += ".pdf"
        plt.savefig(save_pdf, dpi=dpi)

    if display:
        plt.show()



import os
import unittest
from pyspedas.tplot_tools import data_exists, del_data, tplot_names, tplot
import pyspedas
from pyspedas.projects.cluster.load_csa import cl_master_probes, cl_master_datatypes



class LoadTestCases(unittest.TestCase):
    def test_csa(self):
        dtypes = cl_master_datatypes()
        probes = cl_master_probes()
        self.assertTrue('CP_FGM_SPIN' in dtypes)
        self.assertTrue('C1' in probes)

    def test_load_fgm_data(self):
        del_data('*')
        mag_vars = pyspedas.projects.cluster.fgm(time_clip=True)
        self.assertTrue(data_exists('B_xyz_gse__C1_UP_FGM'))
        self.assertTrue('B_xyz_gse__C1_UP_FGM' in mag_vars)

    def test_load_fgm_data_prefix_none(self):
        del_data('*')
        mag_vars = pyspedas.projects.cluster.fgm(prefix=None, time_clip=True)
        self.assertTrue(data_exists('B_xyz_gse__C1_UP_FGM'))
        self.assertTrue('B_xyz_gse__C1_UP_FGM' in mag_vars)

    def test_load_fgm_data_suffix_none(self):
        del_data('*')
        mag_vars = pyspedas.projects.cluster.fgm(suffix=None, time_clip=True)
        self.assertTrue(data_exists('B_xyz_gse__C1_UP_FGM'))
        self.assertTrue('B_xyz_gse__C1_UP_FGM' in mag_vars)

    def test_load_fgm_data_prefix_suffix(self):
        del_data('*')
        mag_vars = pyspedas.projects.cluster.fgm(prefix='pre_', suffix='_suf', time_clip=True)
        self.assertTrue(data_exists('pre_B_xyz_gse__C1_UP_FGM_suf'))
        self.assertTrue('pre_B_xyz_gse__C1_UP_FGM_suf' in mag_vars)

    def test_load_fgm_cp_data(self):
        del_data('*')
        files = pyspedas.projects.cluster.fgm(datatype='cp', trange=['2003-12-15', '2003-12-16'], downloadonly=True)
        self.assertTrue(os.path.exists(files[0]))

    def test_load_asp_data(self):
        del_data('*')
        asp_vars = pyspedas.projects.cluster.aspoc(trange=['2004-04-05', '2004-04-06'])
        self.assertTrue(data_exists('I_ion__C1_PP_ASP'))
        self.assertTrue('I_ion__C1_PP_ASP' in asp_vars)

    def test_load_cis_data(self):
        del_data('*')
        cis_vars = pyspedas.projects.cluster.cis()
        self.assertTrue(data_exists('N_p__C1_PP_CIS'))
        self.assertTrue('N_p__C1_PP_CIS' in cis_vars)

    def test_load_dwp_data(self):
        del_data('*')
        dwp_vars = pyspedas.projects.cluster.dwp()
        self.assertTrue(data_exists('Correl_freq__C1_PP_DWP'))
        self.assertTrue('Correl_freq__C1_PP_DWP' in dwp_vars)

    def test_load_edi_data_download_only(self):
        del_data('*')
        edi_vars = pyspedas.projects.cluster.edi(downloadonly=True)
        self.assertTrue(isinstance(edi_vars, list))

    def test_load_edi_data(self):
        del_data('*')
        edi_vars = pyspedas.projects.cluster.edi()
        self.assertTrue(data_exists('V_ed_xyz_gse__C1_PP_EDI'))
        self.assertTrue('V_ed_xyz_gse__C1_PP_EDI' in edi_vars)

    def test_load_efw_data(self):
        del_data('*')
        efw_vars = pyspedas.projects.cluster.efw()
        self.assertTrue(data_exists('E_pow_f1__C1_PP_EFW'))
        self.assertTrue('E_pow_f1__C1_PP_EFW' in efw_vars)

    def test_load_peace_data(self):
        del_data('*')
        p_vars = pyspedas.projects.cluster.peace()
        self.assertTrue(data_exists('T_e_par__C1_PP_PEA'))
        self.assertTrue('T_e_par__C1_PP_PEA' in p_vars)

    def test_load_rap_data(self):
        del_data('*')
        r_vars = pyspedas.projects.cluster.rapid()
        self.assertTrue(data_exists('J_e_lo__C1_PP_RAP'))
        self.assertTrue('J_e_lo__C1_PP_RAP' in r_vars)

    def test_load_sta_data(self):
        del_data('*')
        sta_vars = pyspedas.projects.cluster.staff()
        self.assertTrue(data_exists('E_pow_f2__C1_PP_STA'))
        self.assertTrue('E_pow_f2__C1_PP_STA' in sta_vars)

    def test_load_wbd_data_notplot(self):
        del_data('*')
        wbd_vars = pyspedas.projects.cluster.wbd(trange=['2012-11-6/02:10', '2012-11-6/02:15'], notplot=True)
        self.assertTrue('WBD_Elec' in wbd_vars)

    def test_load_wbd_data_(self):
        del_data('*')
        wbd_vars = pyspedas.projects.cluster.wbd(trange=['2012-11-6/02:10', '2012-11-6/02:15'])
        self.assertTrue(data_exists('WBD_Elec'))
        self.assertTrue('WBD_Elec' in wbd_vars)

    def test_load_whi_data(self):
        del_data('*')
        whi_vars = pyspedas.projects.cluster.whi()
        self.assertTrue(data_exists('E_pow_f5__C1_PP_WHI'))
        self.assertTrue('E_pow_f5__C1_PP_WHI' in whi_vars)

    def test_load_csa_CE_WBD_WAVEFORM_CDF_data(self):
        del_data('*')
        trange = ['2012-11-06T02:19:00Z', '2012-11-06T02:19:59Z']
        wbd_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CE_WBD_WAVEFORM_CDF'], time_clip=True)
        print(wbd_data)
        self.assertTrue('WBD_Elec' in wbd_data)
        self.assertTrue(data_exists('WBD_Elec'))

    def test_load_csa_CP_AUX_POSGSE_1M_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        pos_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_AUX_POSGSE_1M'], time_clip=True)
        self.assertTrue('sc_r_xyz_gse__C1_CP_AUX_POSGSE_1M' in pos_data)
        self.assertTrue(data_exists('sc_r_xyz_gse__C1_CP_AUX_POSGSE_1M'))

    def test_load_csa_CP_AUX_POSGSE_1M_data_prefix_none(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        pos_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             prefix=None,
                                             datatypes=['CP_AUX_POSGSE_1M'], time_clip=True)
        self.assertTrue('sc_r_xyz_gse__C1_CP_AUX_POSGSE_1M' in pos_data)
        self.assertTrue(data_exists('sc_r_xyz_gse__C1_CP_AUX_POSGSE_1M'))

    def test_load_csa_CP_AUX_POSGSE_1M_data_suffix_none(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        pos_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             suffix=None,
                                             datatypes=['CP_AUX_POSGSE_1M'], time_clip=True)
        self.assertTrue('sc_r_xyz_gse__C1_CP_AUX_POSGSE_1M' in pos_data)
        self.assertTrue(data_exists('sc_r_xyz_gse__C1_CP_AUX_POSGSE_1M'))

    def test_load_csa_CP_AUX_POSGSE_1M_data_prefix_suffix(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        pos_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             prefix='pre_',
                                             suffix='_suf',
                                             datatypes=['CP_AUX_POSGSE_1M'], time_clip=True)
        self.assertTrue('pre_sc_r_xyz_gse__C1_CP_AUX_POSGSE_1M_suf' in pos_data)
        self.assertTrue(data_exists('pre_sc_r_xyz_gse__C1_CP_AUX_POSGSE_1M_suf'))


    def test_load_csa_CP_CIS_CODIF_HS_H1_MOMENTS_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        mom_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_CIS-CODIF_HS_H1_MOMENTS'], time_clip=True)
        print(mom_data)
        self.assertTrue('density__C1_CP_CIS_CODIF_HS_H1_MOMENTS' in mom_data)
        self.assertTrue(data_exists('density__C1_CP_CIS_CODIF_HS_H1_MOMENTS'))

    def test_load_csa_CP_CIS_CODIF_HS_He1_MOMENTS_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        mom_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_CIS-CODIF_HS_He1_MOMENTS'], time_clip=True)
        print(mom_data)
        self.assertTrue('density__C1_CP_CIS_CODIF_HS_He1_MOMENTS' in mom_data)
        self.assertTrue(data_exists('density__C1_CP_CIS_CODIF_HS_He1_MOMENTS'))


    def test_load_csa_CP_CIS_CODIF_HS_O1_MOMENTS_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        mom_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_CIS-CODIF_HS_O1_MOMENTS'], time_clip=True)
        print(mom_data)
        self.assertTrue('density__C1_CP_CIS_CODIF_HS_O1_MOMENTS' in mom_data)
        self.assertTrue(data_exists('density__C1_CP_CIS_CODIF_HS_O1_MOMENTS'))


    def test_load_csa_CP_CIS_CODIF_PAD_HS_H1_PF_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        mom_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_CIS-CODIF_PAD_HS_H1_PF'], time_clip=True)
        print(mom_data)
        self.assertTrue('Differential_Particle_Flux__C1_CP_CIS_CODIF_PAD_HS_H1_PF' in mom_data)
        self.assertTrue(data_exists('Differential_Particle_Flux__C1_CP_CIS_CODIF_PAD_HS_H1_PF'))


    def test_load_csa_CP_CIS_CODIF_PAD_HS_He1_PF_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        mom_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_CIS-CODIF_PAD_HS_He1_PF'], time_clip=True)
        print(mom_data)
        self.assertTrue('Differential_Particle_Flux__C1_CP_CIS_CODIF_PAD_HS_He1_PF' in mom_data)
        self.assertTrue(data_exists('Differential_Particle_Flux__C1_CP_CIS_CODIF_PAD_HS_He1_PF'))


    def test_load_csa_CP_CIS_CODIF_PAD_HS_O1_PF_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        mom_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_CIS-CODIF_PAD_HS_O1_PF'], time_clip=True)
        print(mom_data)
        self.assertTrue('Differential_Particle_Flux__C1_CP_CIS_CODIF_PAD_HS_O1_PF' in mom_data)
        self.assertTrue(data_exists('Differential_Particle_Flux__C1_CP_CIS_CODIF_PAD_HS_O1_PF'))


    def test_load_csa_CP_CIS_HIA_ONBOARD_MOMENTS_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        mom_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_CIS-HIA_ONBOARD_MOMENTS'], time_clip=True)
        print(mom_data)
        self.assertTrue('density__C1_CP_CIS_HIA_ONBOARD_MOMENTS' in mom_data)
        self.assertTrue(data_exists('density__C1_CP_CIS_HIA_ONBOARD_MOMENTS'))

    # This test uses a different time range from test_load_csa_CP_CIS_HIA_ONBOARD_MOMENTS_data,
    # and loads for all four probes.
    def test_load_csa_mom_data(self):
        del_data('*')
        mom_data = pyspedas.projects.cluster.load_csa(probes=['C1', 'C2', 'C3', 'C4'],
                                             trange=['2003-08-17/16:40', '2003-08-17/16:45'],
                                             datatypes=['CP_CIS-HIA_ONBOARD_MOMENTS'], time_clip=True)
        self.assertTrue('density__C1_CP_CIS_HIA_ONBOARD_MOMENTS' in mom_data)
        self.assertTrue(data_exists('density__C1_CP_CIS_HIA_ONBOARD_MOMENTS'))


    def test_load_csa_CP_CIS_HIA_PAD_HS_MAG_IONS_PF_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        mom_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_CIS-HIA_PAD_HS_MAG_IONS_PF'], time_clip=True)
        print(mom_data)
        self.assertTrue('Differential_Particle_Flux__C1_CP_CIS_HIA_PAD_HS_MAG_IONS_PF' in mom_data)
        self.assertTrue(data_exists('Differential_Particle_Flux__C1_CP_CIS_HIA_PAD_HS_MAG_IONS_PF'))

    def test_load_csa_CP_EDI_AEDC_data(self):
        del_data('*')
        trange = ['2005-08-01T00:00:00Z', '2005-08-02T00:00:00Z']
        edi_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_EDI_AEDC'], time_clip=True)
        print(edi_data)
        self.assertTrue('counts_GDU1_PA_90__C1_CP_EDI_AEDC' in edi_data)
        self.assertTrue(data_exists('counts_GDU1_PA_90__C1_CP_EDI_AEDC'))

    def test_load_csa_CP_EDI_MP_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        edi_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_EDI_MP'], time_clip=True)
        self.assertTrue('V_ed_xyz_gse__C1_CP_EDI_MP' in edi_data)
        self.assertTrue(data_exists('V_ed_xyz_gse__C1_CP_EDI_MP'))

    def test_load_csa_CP_EDI_SPIN_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        edi_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_EDI_SPIN'], time_clip=True)
        self.assertTrue('V_ed_xyz_gse__C1_CP_EDI_SPIN' in edi_data)
        self.assertTrue(data_exists('V_ed_xyz_gse__C1_CP_EDI_SPIN'))

    # multidimensional DEPEND_0 array for dsettings/caveat variables
    def test_load_csa_CP_EFW_L2_E3D_INERT_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        efw_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_EFW_L2_E3D_INERT'], time_clip=True)
        self.assertTrue('E_Vec_xyz_ISR2__C1_CP_EFW_L2_E3D_INERT' in efw_data)
        self.assertTrue(data_exists('E_Vec_xyz_ISR2__C1_CP_EFW_L2_E3D_INERT'))

    def test_load_csa_CP_EFW_L2_P_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        efw_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_EFW_L2_P'], time_clip=True)
        self.assertTrue('Spacecraft_potential__C1_CP_EFW_L2_P' in efw_data)
        self.assertTrue(data_exists('Spacecraft_potential__C1_CP_EFW_L2_P'))

    def test_load_csa_CP_EFW_L2_V3D_INERT_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        efw_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_EFW_L2_V3D_INERT'], time_clip=True)
        self.assertTrue('v_drift_ISR2__C1_CP_EFW_L2_V3D_INERT' in efw_data)
        self.assertTrue(data_exists('v_drift_ISR2__C1_CP_EFW_L2_V3D_INERT'))

    def test_load_csa_CP_EFW_L3_E3D_INERT_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        efw_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_EFW_L3_E3D_INERT'], time_clip=True)
        self.assertTrue('E_Vec_xyz_ISR2__C1_CP_EFW_L3_E3D_INERT' in efw_data)
        self.assertTrue(data_exists('E_Vec_xyz_ISR2__C1_CP_EFW_L3_E3D_INERT'))

    def test_load_csa_CP_EFW_L3_P_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        efw_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_EFW_L3_P'], time_clip=True)
        self.assertTrue('Spacecraft_potential__C1_CP_EFW_L3_P' in efw_data)
        self.assertTrue(data_exists('Spacecraft_potential__C1_CP_EFW_L3_P'))

    def test_load_csa_CP_EFW_L3_V3D_INERT_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        efw_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_EFW_L3_V3D_INERT'], time_clip=True)
        self.assertTrue('v_drift_ISR2__C1_CP_EFW_L3_V3D_INERT' in efw_data)
        self.assertTrue(data_exists('v_drift_ISR2__C1_CP_EFW_L3_V3D_INERT'))

    def test_load_csa_CP_FGM_5VPS_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        fgm_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_FGM_5VPS'], time_clip=True)
        self.assertTrue('B_vec_xyz_gse__C1_CP_FGM_5VPS' in fgm_data)
        self.assertTrue(data_exists('B_vec_xyz_gse__C1_CP_FGM_5VPS'))

    def test_load_csa_CP_FGM_FULL_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        fgm_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_FGM_FULL'], time_clip=True)
        self.assertTrue('B_vec_xyz_gse__C1_CP_FGM_FULL' in fgm_data)
        self.assertTrue(data_exists('B_vec_xyz_gse__C1_CP_FGM_FULL'))

    def test_load_csa_CP_FGM_SPIN_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        fgm_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_FGM_SPIN'], time_clip=True)
        self.assertTrue('B_vec_xyz_gse__C1_CP_FGM_SPIN' in fgm_data)
        self.assertTrue(data_exists('B_vec_xyz_gse__C1_CP_FGM_SPIN'))

    def test_load_csa_CP_PEA_MOMENTS_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        mom_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_PEA_MOMENTS'], time_clip=True)
        self.assertTrue('Data_Density__C1_CP_PEA_MOMENTS' in mom_data)
        self.assertTrue(data_exists('Data_Density__C1_CP_PEA_MOMENTS'))

    # multidimensional DEPEND_0 array for dsettings/caveat variables
    def test_load_csa_CP_PEA_PITCH_SPIN_DEFlux_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        pea_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_PEA_PITCH_SPIN_DEFlux'], time_clip=True)
        self.assertTrue('Data__C1_CP_PEA_PITCH_SPIN_DEFlux' in pea_data)
        self.assertTrue(data_exists('Data__C1_CP_PEA_PITCH_SPIN_DEFlux'))

    # multidimensional DEPEND_0 array for dsettings/caveat variables
    def test_load_csa_CP_PEA_PITCH_SPIN_DPFlux_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        pea_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_PEA_PITCH_SPIN_DPFlux'], time_clip=True)
        self.assertTrue('Data__C1_CP_PEA_PITCH_SPIN_DPFlux' in pea_data)
        self.assertTrue(data_exists('Data__C1_CP_PEA_PITCH_SPIN_DPFlux'))

    # multidimensional DEPEND_0 array for dsettings/caveat variables
    def test_load_csa_CP_PEA_PITCH_SPIN_PSD_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        pea_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_PEA_PITCH_SPIN_PSD'], time_clip=True)
        self.assertTrue('Data__C1_CP_PEA_PITCH_SPIN_PSD' in pea_data)
        self.assertTrue(data_exists('Data__C1_CP_PEA_PITCH_SPIN_PSD'))

    def test_load_csa_CP_RAP_ESPCT6_data(self):
        del_data('*')
        trange = ['2008-02-03T00:00:00Z', '2008-02-05T00:00:00Z']
        rap_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_RAP_ESPCT6'], time_clip=True)
        print(rap_data)
        self.assertTrue('Electron_Dif_flux__C1_CP_RAP_ESPCT6' in rap_data)
        self.assertTrue(data_exists('Electron_Dif_flux__C1_CP_RAP_ESPCT6'))

    # multidimensional DEPEND_0 array for dsettings/caveat variables
    def test_load_csa_CP_RAP_ESPCT6_R_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        rap_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_RAP_ESPCT6_R'], time_clip=True)
        self.assertTrue('Electron_Rate__C1_CP_RAP_ESPCT6_R' in rap_data)
        self.assertTrue(data_exists('Electron_Rate__C1_CP_RAP_ESPCT6_R'))

    # multidimensional DEPEND_0 array for dsettings/caveat variables
    def test_load_csa_CP_RAP_HSPCT_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        rap_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_RAP_HSPCT'], time_clip=True)
        self.assertTrue('Proton_Dif_flux__C1_CP_RAP_HSPCT' in rap_data)
        self.assertTrue(data_exists('Proton_Dif_flux__C1_CP_RAP_HSPCT'))

    # multidimensional DEPEND_0 array for dsettings/caveat variables
    def test_load_csa_CP_RAP_HSPCT_R_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        rap_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_RAP_HSPCT_R'], time_clip=True)
        self.assertTrue('Proton_Rate__C1_CP_RAP_HSPCT_R' in rap_data)
        self.assertTrue(data_exists('Proton_Rate__C1_CP_RAP_HSPCT_R'))

    # multidimensional DEPEND_0 array for dsettings/caveat variables
    def test_load_csa_CP_RAP_ISPCT_CNO_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        rap_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_RAP_ISPCT_CNO'], time_clip=True)
        self.assertTrue('CNO_Dif_flux__C1_CP_RAP_ISPCT_CNO' in rap_data)
        self.assertTrue(data_exists('CNO_Dif_flux__C1_CP_RAP_ISPCT_CNO'))

    # multidimensional DEPEND_0 array for dsettings/caveat variables
    def test_load_csa_CP_RAP_ISPCT_He_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        rap_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_RAP_ISPCT_He'], time_clip=True)
        self.assertTrue('Helium_Dif_flux__C1_CP_RAP_ISPCT_He' in rap_data)
        self.assertTrue(data_exists('Helium_Dif_flux__C1_CP_RAP_ISPCT_He'))

    # This date returns data, but it's large enough (tens of megabytes that there's a significant
    # chance of the server not completing the request.
    @unittest.skip
    def test_load_csa_CP_STA_CS_HBR_data(self):
        del_data('*')
        trange = ['2001-01-31T00:00:00Z', '2001-01-31T23:59:59Z']
        sta_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_STA_CS_HBR'], time_clip=True)
        print(sta_data)
        self.assertTrue('Complex_Spectrum__C1_CP_STA_CS_HBR' in sta_data)
        self.assertTrue(data_exists('Complex_Spectrum__C1_CP_STA_CS_HBR'))

    # This date returns data, but it's large enough (tens of megabytes that there's a significant
    # chance of the server not completing the request.
    @unittest.skip
    def test_load_csa_CP_STA_CS_NBR_data(self):
        del_data('*')
        trange = ['2001-01-31T00:00:00Z', '2001-01-31T23:59:59Z']
        sta_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_STA_CS_NBR'], time_clip=True)
        print(sta_data)
        self.assertTrue('Complex_Spectrum__C1_CP_STA_CS_NBR' in sta_data)
        self.assertTrue(data_exists('Complex_Spectrum__C1_CP_STA_CS_NBR'))

    def test_load_csa_CP_STA_CWF_GSE_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        sta_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_STA_CWF_GSE'], time_clip=True)
        self.assertTrue('B_vec_xyz_Instrument__C1_CP_STA_CWF_GSE' in sta_data)
        self.assertTrue(data_exists('B_vec_xyz_Instrument__C1_CP_STA_CWF_GSE'))

    # 404 error, no data?
    @unittest.skip
    def test_load_csa_CP_STA_CWF_HBR_ISR2_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        sta_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_STA_CWF_HBR_ISR2'], time_clip=True)
        print(sta_data)
        self.assertTrue('B_vec_xyz_Instrument__C1_CP_STA_CWF_GSE' in sta_data)
        self.assertTrue(data_exists('B_vec_xyz_Instrument__C1_CP_STA_CWF_GSE'))

    # 404 error, no data?
    @unittest.skip
    def test_load_csa_CP_STA_CWF_NBR_ISR2_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        sta_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_STA_CWF_NBR_ISR2'], time_clip=True)
        print(sta_data)
        self.assertTrue('B_vec_xyz_Instrument__C1_CP_STA_CWF_GSE' in sta_data)
        self.assertTrue(data_exists('B_vec_xyz_Instrument__C1_CP_STA_CWF_GSE'))

    # Spectrogram variables have string-valued DEPEND_2, which is not ISTP compliant.
    def test_load_csa_CP_STA_PSD_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        sta_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_STA_PSD'], time_clip=True)
        print(sta_data)
        #self.assertTrue('B_vec_xyz_Instrument__C1_CP_STA_CWF_GSE' in sta_data)
        #self.assertTrue(data_exists('B_vec_xyz_Instrument__C1_CP_STA_CWF_GSE'))

    def test_load_csa_CP_WBD_WAVEFORM_data(self):
        del_data('*')
        trange = ['2001-02-04T13:40:00Z', '2001-02-04T13:49:59Z']
        wbd_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_WBD_WAVEFORM'], time_clip=True)
        print(wbd_data)
        self.assertTrue('E__C1_CP_WBD_WAVEFORM' in wbd_data)
        self.assertTrue(data_exists('E__C1_CP_WBD_WAVEFORM'))

    def test_load_csa_CP_WHI_ELECTRON_DENSITY_data(self):
        del_data('*')
        trange = ['2018-01-12T00:00:00Z', '2018-01-13T00:00:00Z']
        whi_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_WHI_ELECTRON_DENSITY'], time_clip=True)
        print(whi_data)
        self.assertTrue('Electron_Density__C1_CP_WHI_ELECTRON_DENSITY' in whi_data)
        self.assertTrue(data_exists('Electron_Density__C1_CP_WHI_ELECTRON_DENSITY'))

    def test_load_csa_CP_CP_WHI_NATURAL_data(self):
        del_data('*')
        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        whi_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                             trange=trange,
                                             datatypes=['CP_WHI_NATURAL'], time_clip=True)
        self.assertTrue('Electric_Spectral_Power_Density__C1_CP_WHI_NATURAL' in whi_data)
        self.assertTrue(data_exists('Electric_Spectral_Power_Density__C1_CP_WHI_NATURAL'))

    def test_load_csa_JP_AUX_PMP_data(self):
        del_data('*')
        trange = ['2017-01-01T00:00:00Z', '2017-02-01T00:00:00Z']
        jp_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                            trange=trange,
                                            datatypes=['JP_AUX_PMP'], time_clip=True)
        print(jp_data)
        self.assertTrue('L_value__C1_JP_AUX_PMP' in jp_data)
        self.assertTrue(data_exists('L_value__C1_JP_AUX_PMP'))

    def test_load_csa_JP_PSE_data(self):
        del_data('*')
        trange = ['2017-01-01T00:00:00Z', '2018-01-02T00:00:00Z']
        jp_data = pyspedas.projects.cluster.load_csa(probes=['C1'],
                                            trange=trange,
                                            datatypes=['JP_AUX_PSE'], time_clip=True)
        print(jp_data)
        self.assertTrue('sc_r1_xyz_gse__C1_JP_AUX_PSE' in jp_data)
        self.assertTrue(data_exists('sc_r1_xyz_gse__C1_JP_AUX_PSE'))


if __name__ == '__main__':
    unittest.main()



import unittest
import os
from pyspedas import mms_overview_plot

global_display = False  # Set to False for Github testing

class TestMMSOverviewPlot(unittest.TestCase):
    def test_default_overview_plot(self):
        save_png='mms_default_overview'
        mms_overview_plot(display=global_display,save_png=save_png)
        self.assertTrue(os.path.exists(save_png + '.png'))

if __name__ == '__main__':
    unittest.main()



from .load import load

# This routine was originally in stereo/__init__.py, until being moved to its own file.
# Please refer to __init__.py if you need to see the revision history before it was moved.

def plastic(trange=['2013-11-5', '2013-11-6'],
            probe='a',
            datatype='1min',
            level='l2',
            suffix='',
            prefix='',
            get_support_data=False,
            varformat=None,
            varnames=[],
            downloadonly=False,
            notplot=False,
            no_update=False,
            time_clip=False):
    """
    This function loads data from the PLASTIC instrument

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2013-11-5', '2013-11-6']

        probe: str
            Spacecraft probe ('a' for ahead, 'b' for behind)
            Default: 'a'

        datatype: str
            Data type; Valid options: 1min
            Default: '1min'

        suffix: str
            The tplot variable names will be given this suffix.
            Default: '', no suffix is added.

        prefix: str
            The tplot variable names will be given this prefix.
            Default: '', no prefix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.
            Default: False. Only loads in data with a "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.
            Default: None. All variables are loaded in.

        varnames: list of str
            List of variable names to load.
            Default: [], if not specified, all data variables are loaded

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

    Returns
    ----------
        List of tplot variables created.

    Example:
    ----------
        >>> import pyspedas
        >>> from pyspedas import tplot
        >>> plastic_vars = pyspedas.projects.stereo.plastic(trange=['2013-11-5', '2013-11-6'])
        >>> tplot(['proton_number_density', 'proton_bulk_speed', 'proton_temperature', 'proton_thermal_speed'])

    """
    return load(instrument='plastic', trange=trange, probe=probe, level=level, datatype=datatype, suffix=suffix, prefix=prefix,
                get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly,
                notplot=notplot, time_clip=time_clip, no_update=no_update)



import pyspedas
import numpy as np
import copy
import logging

def avg_res_data(tvar,res,newname=None):
    """
    Averages the variable over a specified period of time.

    Parameters
    ----------
        tvar : str
            Name of tplot variable.
        res : int/float
            The new data resolution
        newname : str
            Name of new tvar for averaged data.  If not set, then the data in tvar is replaced.

    Returns
    -------
        None

    Note
    ----

    This routine uses the xarray coarsen() method to reduce the time resolution.  It will only work if the
    data is evenly gridded, and the res parameter evenly divides the number of samples.
    For most purposes, it is more appropriate to use pyspedas.avg_data() instead.

    Examples
    --------

        >>> #Average the data over every two seconds
        >>> pyspedas.store_data('d', data={'x':[2,5,8,11,14,17,21], 'y':[[1,1,50],[2,2,3],[100,4,47],[4,90,5],[5,5,99],[6,6,25],[7,7,-5]]})
        >>> pyspedas.avg_res_data('d',2,'d2res')

    """

    tvar_new = pyspedas.tplot_tools.data_quants[tvar].coarsen(time=res, boundary='trim').mean()
    tvar_new.name = pyspedas.tplot_tools.data_quants[tvar].name
    tvar_new.attrs = copy.deepcopy(pyspedas.tplot_tools.data_quants[tvar].attrs)

    if newname is None:
        pyspedas.tplot_tools.data_quants[tvar] = newname
    else:
        if 'spec_bins' in pyspedas.tplot_tools.data_quants[tvar].coords:
           pyspedas.store_data(newname, data={'x': tvar_new.coords['time'].values, 'y': tvar_new.values,
                                               'v': tvar_new.coords['spec_bins'].values})
        else:
           pyspedas.store_data(newname, data={'x': tvar_new.coords['time'].values, 'y': tvar_new.values})

        pyspedas.tplot_tools.data_quants[newname].attrs = copy.deepcopy(pyspedas.tplot_tools.data_quants[tvar].attrs)

    return




"""
Load data from the Cluster Science Archive.

This loading function uses the Cluster Science Archive:
    https://csa.esac.esa.int/
It is a web service, we create the query and the web service responds
with a CDF file which is packaged as tar.gz.

We download the tar.gz file directly, without using pyspedas.download().
"""
import logging
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import time_string
from pyspedas.tplot_tools import time_double
from pyspedas.tplot_tools import cdf_to_tplot

import requests
import sys
import tarfile
import os
from pathlib import Path
from typing import List
from .config import CONFIG

from pyspedas.utilities.download import is_fsspec_uri
import fsspec

def cl_master_datatypes():
    """Return list of data types."""
    md = ['CE_WBD_WAVEFORM_CDF', 'CP_AUX_POSGSE_1M',
          'CP_CIS-CODIF_HS_H1_MOMENTS', 'CP_CIS-CODIF_HS_He1_MOMENTS',
          'CP_CIS-CODIF_HS_O1_MOMENTS', 'CP_CIS-CODIF_PAD_HS_H1_PF',
          'CP_CIS-CODIF_PAD_HS_He1_PF', 'CP_CIS-CODIF_PAD_HS_O1_PF',
          'CP_CIS-HIA_ONBOARD_MOMENTS', 'CP_CIS-HIA_PAD_HS_MAG_IONS_PF',
          'CP_EDI_AEDC', 'CP_EDI_MP', 'CP_EDI_SPIN', 'CP_EFW_L2_E3D_INERT',
          'CP_EFW_L2_P', 'CP_EFW_L2_V3D_INERT', 'CP_EFW_L3_E3D_INERT',
          'CP_EFW_L3_P', 'CP_EFW_L3_V3D_INERT', 'CP_FGM_5VPS', 'CP_FGM_FULL',
          'CP_FGM_SPIN', 'CP_PEA_MOMENTS', 'CP_PEA_PITCH_SPIN_DEFlux',
          'CP_PEA_PITCH_SPIN_DPFlux', 'CP_PEA_PITCH_SPIN_PSD', 'CP_RAP_ESPCT6',
          'CP_RAP_ESPCT6_R', 'CP_RAP_HSPCT', 'CP_RAP_HSPCT_R',
          'CP_RAP_ISPCT_CNO', 'CP_RAP_ISPCT_He', 'CP_STA_CS_HBR',
          'CP_STA_CS_NBR', 'CP_STA_CWF_GSE', 'CP_STA_CWF_HBR_ISR2',
          'CP_STA_CWF_NBR_ISR2', 'CP_STA_PSD', 'CP_WBD_WAVEFORM',
          'CP_WHI_ELECTRON_DENSITY', 'CP_WHI_NATURAL', 'JP_AUX_PMP', 'JP_AUX_PSE']
    return md


def cl_master_probes():
    """Return list of probe names."""
    mp = ['C1', 'C2', 'C3', 'C4']
    return mp


def cl_format_time(s):
    """Return a string formated for Cluster web services."""
    # Date format: YYYY-MM-DDThh:mm:ssZ
    r = time_string(time_double(s), "%Y-%m-%dT%H:%M:%SZ")
    return r


def load_csa(trange:List[str]=['2001-02-01', '2001-02-03'],
             probes:List[str]=['C1'],
             datatypes:List[str]=['CP_CIS-CODIF_HS_H1_MOMENTS'],
             downloadonly:bool=False,
             time_clip:bool=True,
             prefix:str='',
             suffix:str='',
             get_support_data:bool=False,
             varformat:str=None,
             varnames:List[str]=[],
             notplot:bool=False) -> List[str]:
    """Load data using the Cluster Science Data archive.

    Parameters
    ----------
        trange : list of str
            Time range [start, end].
            Default: ['2001-02-01', '2001-02-03']

        probes : list of str
            List of Cluster probes. Valid options: 'C1','C2','C3','C4', '*' to load all probes
            Default: ['C1']

        datatypes : list of str
            List of Cluster data types. Valid options::
              'CE_WBD_WAVEFORM_CDF', 'CP_AUX_POSGSE_1M',
              'CP_CIS-CODIF_HS_H1_MOMENTS', 'CP_CIS-CODIF_HS_He1_MOMENTS',
              'CP_CIS-CODIF_HS_O1_MOMENTS', 'CP_CIS-CODIF_PAD_HS_H1_PF',
              'CP_CIS-CODIF_PAD_HS_He1_PF', 'CP_CIS-CODIF_PAD_HS_O1_PF',
              'CP_CIS-HIA_ONBOARD_MOMENTS', 'CP_CIS-HIA_PAD_HS_MAG_IONS_PF',
              'CP_EDI_AEDC', 'CP_EDI_MP', 'CP_EDI_SPIN', 'CP_EFW_L2_E3D_INERT',
              'CP_EFW_L2_P', 'CP_EFW_L2_V3D_INERT', 'CP_EFW_L3_E3D_INERT',
              'CP_EFW_L3_P', 'CP_EFW_L3_V3D_INERT', 'CP_FGM_5VPS', 'CP_FGM_FULL',
              'CP_FGM_SPIN', 'CP_PEA_MOMENTS', 'CP_PEA_PITCH_SPIN_DEFlux',
              'CP_PEA_PITCH_SPIN_DPFlux', 'CP_PEA_PITCH_SPIN_PSD', 'CP_RAP_ESPCT6',
              'CP_RAP_ESPCT6_R', 'CP_RAP_HSPCT', 'CP_RAP_HSPCT_R',
              'CP_RAP_ISPCT_CNO', 'CP_RAP_ISPCT_He', 'CP_STA_CS_HBR',
              'CP_STA_CS_NBR', 'CP_STA_CWF_GSE', 'CP_STA_CWF_HBR_ISR2',
              'CP_STA_CWF_NBR_ISR2', 'CP_STA_PSD', 'CP_WBD_WAVEFORM',
              'CP_WHI_ELECTRON_DENSITY', 'CP_WHI_NATURAL', 'JP_AUX_PMP', 'JP_AUX_PSE'

            Default: ['CP_CIS-CODIF_HS_H1_MOMENTS']

        downloadonly: bool
            If true, do not use cdf_to_tplot.
            Default: False

        time_clip: bool
            If true, apply time clip to data.
            Default: False

        suffix: str
            The tplot variable names will be given this suffix.
            Default: ''

        get_support_data: bool
            If True, data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.
            Default: False

        varformat : str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.
            Default: None (all variables will be loaded)

        varnames: str or list of str
            Load these variables only. If [] or ['*'], then load everything.
            Default: []

        notplot: bool
            If True, then data are returned in a hash table instead of
            being stored in tplot variables (useful for debugging, and
            access to multi-dimensional data products)
            Default: False

    Returns
    -------
        list of str
            List of tplot variables created (unless notplot keyword is used).

    Examples
    --------

    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> fgm_vars = pyspedas.projects.cluster.load_csa(trange=['2008-11-01','2008-11-02'],datatypes=['CP_FGM_FULL'])
    >>> tplot(['B_vec_xyz_gse__C1_CP_FGM_FULL','B_mag__C1_CP_FGM_FULL'])
    """
    # Empty output in case of errors.
    tvars = []

    if prefix is None:
        prefix = ''
    if suffix is None:
        suffix = ''

    # Start and end dates
    start_date = cl_format_time(trange[0])
    end_date = cl_format_time(trange[1])

    # Delivery format
    delivery_format = 'CDF_ISTP'
    # Delivery interval
    delivery_interval = 'ALL'

    if not probes:
        return tvars

    if not datatypes:
        return tvars

    if not isinstance(probes, list):
        probes = [probes]

    if not isinstance(datatypes, list):
        datatypes = [datatypes]

    # TODO: Create a function that can resolve wildcards
    # similar to IDL spedas ssl_check_valid_name
    # my_datatypes=ssl_check_valid_name(uc_datatypes,master_datatypes)
    # my_probes=ssl_check_valid_name(uc_probes,master_probes)
    if probes[0] == '*':  # load all probes
        probes = cl_master_probes()

    # Construct the query string
    base_url = 'https://csa.esac.esa.int/csa-sl-tap/data?'
    query_string = ('retrieval_type=PRODUCT&START_DATE=' + start_date +
                    '&END_DATE=' + end_date +
                    '&DELIVERY_FORMAT=' + delivery_format +
                    '&DELIVERY_INTERVAL=' + delivery_interval +
                    '&NON_BROWSER')

    for p in probes:
        for d in datatypes:
            query_string += '&DATASET_ID=' + p + '_' + d

    # Encode the url urllib.parse.quote
    url = base_url + (query_string)

    local_path = CONFIG['local_data_dir'] # could be URI
    if is_fsspec_uri(local_path):
        local_protocol, lpath = local_path.split("://")
        local_fs = fsspec.filesystem(local_protocol, anon=False)

        out_gz = '/'.join([local_path, 'temp_cluster_file.tar.gz'])  # Temp file name
        fileobj = local_fs.open(out_gz, 'wb')
    else:
        Path(local_path).mkdir(parents=True, exist_ok=True)
        out_gz = os.path.join(local_path, 'temp_cluster_file.tar.gz')  # Temp file name
        fileobj = open(out_gz, 'wb')

    # Download the file.
    logging.info("Downloading Cluster data, please wait....")
    try:
        r = requests.get(url, allow_redirects=True)
        r.raise_for_status()
    except requests.exceptions.HTTPError as err:
        logging.error("Download HTTP error: " + str(err))
        return tvars
    except requests.exceptions.RequestException as e:
        logging.error("Download error: " + str(e))
        return tvars
    logging.info("Download complete.")

    # Open the downloaded file.
    with fileobj as w:
        w.write(r.content)

    # Extract the tar archive.
    if is_fsspec_uri(out_gz):
        # Cloud-Awareness: Opens byte stream for tarfile package.
        bo = local_fs.open(out_gz, "rb")
        tar = tarfile.open(fileobj=bo)
    else:
        tar = tarfile.open(out_gz, "r:gz")
    f = tar.getnames()

    for member in tar.getmembers():
        if member.isfile():
            p = '/'.join([local_path, member.path])
            if is_fsspec_uri(p):
                membo = local_fs.open(p, "wb")
            else:
                os.makedirs(str(Path(p).parent), exist_ok=True)
                membo = open(p, "wb")

            # Python > 3.9 requirement from setup.py
            # note: data is written after file is read into memory
            # https://stackoverflow.com/a/62247729
            with tar.extractfile(member.path) as tarbo:
                membo.write(tarbo.read())

            membo.close()
    tar.close()
    # Remove the tar.gz file but keep the extracted.
    if is_fsspec_uri(out_gz):
        local_fs.delete(out_gz)
    else:
        os.remove(out_gz)

    # Get unique set of files.
    f_set = set(f)
    # File list with full path.
    sep = "/" if is_fsspec_uri(local_path) else os.path.sep
    out_files = [sep.join([local_path, s]) for s in list(f_set)]
    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    # Load data into tplot
    tvars = cdf_to_tplot(out_files,
                         prefix=prefix,
                         suffix=suffix,
                         get_support_data=get_support_data,
                         varformat=varformat,
                         varnames=varnames,
                         notplot=notplot)

    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars



import pyspedas
import pyspedas
import numpy as np
from scipy.interpolate import interp1d
from pyspedas.tplot_tools.time_double import time_double

def dens_pot(scpot, offset, pder_calc=False):
    """
    Evaluates McFadden's empirical formula for calculating density as a function of spacecraft potiential.

    Parameters
    ----------
    scpot : list or float
        Spacecraft potential.
    offset : float
        An offset.
    pder_calc : bool=True
        If True, the partial derivative with respect to `scpot` will be calculated.

    Returns
    -------
    result : list or float
        The calculated result based on the empirical formula.
    pder : list of float, optional
        The partial derivative with respect to `scpot`, if `pder` was provided.

    Examples
    --------
    >>> dens_pot(-5.0, 10.0)
    Calculated density potential based on the spacecraft potential and offset.

    >>> dens_pot(-5.0, 10.0, True)
    (Tuple containing the calculated density potential and its partial derivative with respect to the spacecraft potential.)

    References
    ----------
    Based on IDL function THM_SCPOT2DENS.PRO by M.Feuerstein, 5/18/2009.
    SPEDAS/projects/themis/common/thm_scpot2dens.pro
    """

    delta = offset - scpot
    result = 460. * np.power(10, delta / 1.5) + 34. * np.power(10, delta / 0.7) + 1.6 * np.power(10, delta / 30.)

    if pder_calc:
        pder = np.log(10) * ((460. / 1.5) * np.power(10, delta / 1.5) + (34. / 0.7) * np.power(10, delta / 0.7) + (
                    1.6 / 30.) * np.power(10, delta / 30.))
        return result, pder

    return result


def interpolate_to_scptime(source_time, target_time, source_data):
    """
    Interpolates source data to the target time base, handling non-finite values.

    This function takes a series of data points defined at certain times (`source_data` at `source_time`)
    and interpolates these data points to a new series of times (`target_time`). It specifically handles
    cases where the source data might contain non-finite values (e.g., NaN or infinity) by filtering those
    out before interpolation. If there are not enough finite data points to perform interpolation, it returns
    an array of NaN values of the same shape as `target_time`.

    Parameters
    ----------
    source_time : array_like
        The time points corresponding to the `source_data` values.
    target_time : array_like
        The time points to which `source_data` should be interpolated.
    source_data : array_like
        The data values that need to be interpolated to the new time base.

    Returns
    -------
    ndarray
        The interpolated data values at `target_time`. If interpolation is not possible due to insufficient
        finite data points in `source_data`, returns an array of NaN values matching the shape of `target_time`.

    Examples
    --------
    >>> source_time = [1, 2, 3, 4]
    >>> source_data = [1, np.nan, 3, 4]
    >>> target_time = [2, 3]
    >>> interpolate_to_scptime(source_time, target_time, source_data)
    array([2., 3.5])

    Notes
    -----
        - This function is not intended for extenal use, but could be usefull for other analytical tools.
    """

    # Ensure source_time and source_data are numpy arrays for boolean indexing
    source_time = np.array(source_time)
    source_data = np.array(source_data)

    # Filter out non-finite values to avoid interpolation errors
    finite_indices = np.isfinite(source_data)
    if np.sum(finite_indices) < 2:
        # Not enough data points to interpolate
        return np.full_like(target_time, np.nan)

    interp_func = interp1d(source_time[finite_indices], source_data[finite_indices],
                           bounds_error=False, fill_value=np.nan)

    return interp_func(target_time)


def scpot2dens(scpot, scptime, Te, Tetime, dens_e, dens_e_time, dens_i, dens_i_time, probe='c'):
    """
    Calculates spacecraft potential derived density.

    This function processes spacecraft potential data (`scpot`), alongside interpolated electron temperature (`Te`),
    electron density (`dens_e`), and ion density (`dens_i`) data, to calculate electron and ion densities at specific
    time points (`scptime`). It interpolates `Te`, `dens_e`, and `dens_i` data to align with the `scpot` time base,
    It accounts for specific bias changes and offset matrices.

    Parameters
    ----------
    scpot : array_like
        Spacecraft potential data, expected as a 1-D array. (call pyspedas.projects.themis.esa, DATATYPE = 'peer_sc_pot')
    scptime : array_like
        The time base of `scpot`.
    Te : array_like
        Electron temperature data, expected as a 1-D array. (call pyspedas.projects.themis.esa, DATATYPE = 'peer_avgtemp')
    Tetime : array_like
        The time base of `Te`.
    dens_e : array_like
        Electron density data, expected as a 1-D array. (call pyspedas.projects.themis.esa, DATATYPE = 'peer_density')
    dens_e_time : array_like
        The time base for `dens_e`.
    dens_i : array_like
        Ion density data, expected as a 1-D array. (call pyspedas.projects.themis.esa, DATATYPE = 'peir_density')
    dens_i_time : array_like
        The time base for `dens_i`.
    probe_in : int or str, default='c'
        Identifier for the probe being used for measurements. Valid values are 'a', 'b', 'c', 'd', 'e'.

    Returns
    -------
    dens : ndarray
        The calculated densities at `scpot` time points. Returns None if the probe is invalid, if `dens_e` and `dens_i`
        are either not present or do not have at least two elements (indicating insufficient data for interpolation),
        or if the densities remain unmodified due to a lack of valid data.

    Notes
    -----
    - Incorporates probe-specific adjustments through bias changes and offset matrices, applying different
      calculation methodologies across distinct plasma conditions and time periods.
    - Invalid inputs or insufficient density data lead to an early termination of the function with None as return.

    Examples
    --------
    Given a specific time range, the following example demonstrates how to load THEMIS spacecraft data for probe 'c',
    extract density, spacecraft potential, and temperature data, then calculate the plasma density from spacecraft potential
    using the `scpot2dens` function.



    >>> import pyspedas
    >>> import pyspedas
    >>> from pyspedas.projects.themis import scpot2dens
    >>>
    >>> # Define the time range for which to load the data
    >>> trange = ['2007-7-20/17:00:00', '2007-7-20/17:20:00']

    >>> # Load THEMIS ESA data for probe 'c' within the specified time range
    >>> pyspedas.projects.themis.esa(trange=trange,
    >>>                    probe='c',
    >>>                    varnames=['thc_peer_density', 'thc_peir_density', 'thc_peer_sc_pot', 'thc_peer_avgtemp'],
    >>>                    level='l2')
    >>>
    >>> # Retrieve electron density
    >>> dens_e_time, dens_e = get_data('thc_peer_density')
    >>>
    >>> # Retrieve ion density
    >>> dens_i_time, dens_i = get_data('thc_peir_density')
    >>>
    >>> # Retrieve spacecraft potential
    >>> sc_pot_time, sc_pot = get_data('thc_peer_sc_pot')
    >>>
    >>> # Retrieve electron temperature
    >>> Te_time, Te = get_data('thc_peer_avgtemp')
    >>>
    >>> # Calculate the plasma density from spacecraft potential
    >>> Npot = scpot2dens(sc_pot, sc_pot_time, Te, Te_time, dens_e, dens_e_time, dens_i, dens_i_time, 'c')


    References
    ----------
    Based on IDL function THM_SCPOT2DENS.PRO by W.M.Feuerstein, 2009-05-18, with updates and bug fixes in subsequent versions.
    SPEDAS/projects/themis/common/thm_scpot2dens.pro
    """

    # Probe validation and selection
    valid_probes = ['a', 'b', 'c', 'd', 'e']

    if str(probe).lower() not in valid_probes:
        pyspedas.logger.error("Invalid probe entered. Valid probes are [a, b, c, d, e]")
        return None

    # Check if Ne and Ni exist and have >= 2 elements
    if not (dens_e[0] != -1 and dens_i[0] != -1) or (len(dens_e) < 2 or len(dens_i) < 2):
        pyspedas.logger.error("dens_e and dens_i must exist and have two or more elements.")
        return None

    # Interpolate Te, dens_e, and dens_i to scpot time base
    Teint = interpolate_to_scptime(Tetime, scptime, Te)
    dens_e_int = interpolate_to_scptime(dens_e_time, scptime, dens_e)
    dens_i_int = interpolate_to_scptime(dens_i_time, scptime, dens_i)

    # Common bias change and offset matrix for probes 'a', 'b', 'd', and 'e'
    common_bias_change = [time_double('2001-1-1')]
    common_offset_matrix = np.array([[2.2, 2.0, np.nan, np.nan]])

    # Define bias changes and offset matrices for each probe with more compact representation
    bias_changes = {
        probe: {
            'biaschange': common_bias_change,
            'offsetmatrix': common_offset_matrix
        } for probe in ['a', 'b', 'd', 'e']
    }

    # Adding the unique entry for probe 'c'
    bias_changes['c'] = {
        'biaschange': [time_double('2001-1-1'),
                       time_double('2007-7-20 17:24')],
        'offsetmatrix': np.array([[2.2, 2.0, np.nan, np.nan],
                                  [2.9, 2.0, np.nan, np.nan]])
    }

    biaschange_for_probe = bias_changes[probe]['biaschange']
    offsetmatrix_for_probe = bias_changes[probe]['offsetmatrix']

    # Interpolate Te to scpot time base
    Te_interp_func = interp1d(Tetime, Te, bounds_error=False, fill_value="extrapolate")
    Teint = Te_interp_func(scptime)

    # Initialize density array
    dens = np.full_like(scpot, np.nan)

    # Process each bias period
    for i, bias_time in enumerate(biaschange_for_probe):
        # Define conditions for different plasma regions and bias periods
        if i != len(biaschange_for_probe) - 1:
            next_bias = biaschange_for_probe[i + 1]
        else:
            next_bias = time_double('9999-12-31')  # A future date to ensure the last period is included

        time_index = (scptime >= biaschange_for_probe[i]) & (scptime < next_bias)
        if not np.any(time_index):
            continue

        wll = (scpot <= 6.0) & (Teint <= 500) & time_index
        wlh = (scpot <= 6.0) & (Teint > 500) & time_index
        whl = (scpot > 6.0) & (Teint <= 500) & time_index
        whh = (scpot > 6.0) & (Teint > 500) & time_index
        wnf = (~np.isfinite(scpot) | ~np.isfinite(Teint)) & time_index

        # Apply density calculations based on conditions
        # Assuming dens_pot function is defined elsewhere and calculates density based on spacecraft potential and offset
        if np.any(wll):
            dens[wll] = dens_pot(scpot[wll], offsetmatrix_for_probe[i, 0])  # Example calculation, adjust as needed
        if np.any(wlh):
            dens[wlh] = dens_pot(scpot[wlh], offsetmatrix_for_probe[i, 1])  # Adjust as needed

        # For whl and whh, using a simplified formula directly as an example
        if np.any(whl):
            dens[whl] = np.exp((-scpot[whl] + 12) / ((scpot[whl] * 0.14) + 3.36)) / np.sqrt(Teint[whl])
            # Recompute for cold electrons if needed, similar logic to IDL's idxT
        if np.any(whh):
            dens[whh] = np.exp((-scpot[whh] + 12) / ((scpot[whh] * 0.14) + 3.36)) / np.sqrt(Teint[whh])
            # Recompute for cold electrons if needed

        # Handling non-finite values
        if np.any(wnf):
            dens[wnf] = np.nan  # Assign NaN for non-finite conditions

    return dens



import numpy as np
from scipy.integrate import RK45
from geopack import geopack, t89

def trace_to_iono(f, pos_init, max_distance=50.0, dt=1.0, south=False):
    """
    Trace the magnetic field line using RK45 until the ionosphere (defined as 1.1 earth radii) is reached

    Parameters:
    - f: Function that returns the normalized magnetic field direction.
    - pos_init: Initial position [x, y, z].
    - max_distance: Maximum distance to trace before stopping.
    - dt: Initial time step for RK45.

    Returns:
    - positions: List of positions along the traced field line.
    """

    # Initialize the RK45 integrator
    # atol = .0008 Re corresponds to an error tolerance of about 5 km
    # debug: set dt to np.inf for now to see what happens
    dt = 0.05
    integrator = RK45(f, 0, pos_init, 10.0, atol=.0008, rtol=2.0e-10, max_step=dt)

    #integrator.direction = -1.0

    # Store the initial position
    positions = [pos_init]
    radius_last=100.0
    while integrator.t < max_distance:
        # Integrate one step
        integrator.step()

        # Get the current position
        pos_current = integrator.y

        radius = np.linalg.norm(pos_current)

        # Store the current position
        positions.append(pos_current)
        print(pos_current, radius)
        radius_last=radius

        # Check if we've reached the ionosphere (1 Re plus 100 km)
        if radius < 1.015696:
            # We're done
            # Use the last two (or three?) points to refine the foot point, then replace last position with foot point
            break


    return np.array(positions)

def trace_iono_89(time, startpos, iopt=3.0, south=False):
    ps = geopack.recalc(time)
    if south:
        direction=-1.0
    else:
        direction=1.0

    def t89_closure(t,pos):
        b_igrf = geopack.igrf_gsm(pos[0],pos[1],pos[2])
        b_t89 = t89.t89(iopt,ps, pos[0],pos[1],pos[2])
        b = np.array(b_igrf)+np.array(b_t89)
        nb = b/np.linalg.norm(b)
        #print(t,pos,b, nb)
        return direction*nb
    trace_points = trace_to_iono(t89_closure,pos_init=startpos,max_distance=100.0,dt=0.05, south=south)
    foot_point = trace_points[-1]
    print(foot_point)




import pyspedas
from pyspedas.tplot_tools import store_data, get_data, tnames
import copy
import numpy as np
import logging


def deflag(tvar, flag=None, newname=None, method=None, fillval=None):
    """
    Replace NaN or other 'flag' values in arrays with interpolated or other values.

    Parameters
    ----------

        tvar: str or list[str]
            Names of tplot variables to deflag (wildcards accepted)
        flag : int,list
            Flagged data will be converted to NaNs.
        method : str, optional
            Method to apply. Valid options::

                'repeat': Repeat last good value
                'linear': Interpolate linearly over gap
                'replace': Replace flagged values with fillval, or NaN if fillval not specified
                'remove_nan': Remove timestamps and values with a NaN in any dimension

        newname : str
            Name of new tvar for deflagged data storage.
            If not specified, then the data in tvar1 will be replaced.
            This is not an option for multiple variable input, for
            multiple or pseudo variables, the data is overwritten.
        fillval: int, float (optional)
            Value to use as replacement when method='replace'

    Notes
    -----

       deflag only works for 1 or 2-d data arrays; ntimes or (ntimes, nspectral_bins)

    Returns
    -------

        None

    Examples
    --------

        >>> pyspedas.store_data('d', data={'x':[2,5,8,11,14,17,21], 'y':[[1,1],[2,2],[100,4],[4,90],[5,5],[6,6],[7,7]]})
        >>> # Remove any instances of [100,90,7,2,57] from 'd', store in 'e'.
        >>> pyspedas.deflag('d',[100,90,7,2,57],newname='e')

    """

    # for linear method, and flag of NaN, or none, interp_nan an be called
    #    if (flag == None or np.isnan(flag)) and method == 'linear':
    #        interp_nan(tvar, newname=newname)
    #        return
    
    # check for globbed or array input, and call recursively
    tn = tnames(tvar)
    if len(tn) == 0:
        return
    elif len(tn) > 1:
        for j in range(len(tn)):
            deflag(tn[j], flag, method=method, fillval=fillval)
        return

    # Now flag needs to be an array
    if flag is None:
        flag = np.zeros(1)+np.nan
    else:
        if not isinstance(flag, np.ndarray):
            flag_array = np.array(flag)
        else:
            flag_array = flag
        # Ok, now if flag was a scalar, the array will have ndim = 0
        if flag_array.ndim == 0:
            flag = np.array([flag])
        else:
            flag = flag_array
    
    nf = len(flag)
    if method == 'remove_nan':  # this is different from the other methods, which retain all time intervals
        a = copy.deepcopy(get_data(tvar))
        alen = len(a)
        # Ignore more than 2d Y input
        if alen > 3:
            logging.info('deflag is not used for more than 2-d input')
            return

        time = a[0]
        data = a[1]
        # Added this to prevent error on !-d arrays (see below)
        data_dim = data.ndim
        new_time = []
        new_data = []
        
        if alen == 3:  # v variable
            v = a[2]
            if v.ndim == 1:
                new_v = v
                append_v = False
            else:
                new_v = []
                append_v = True

        # Fill the new variable
        non_nan_seen = False
        for j in range(len(time)):
            # This used to be "if len(data[j]) > 1", which failed if data is 1-D so that data[j] is scalar
            # Instead we go by the dimensions of the data array itself
            if data_dim > 1:
                tj = np.sum(data[j])
            else:
                tj = data[j]
            if not np.isnan(tj):
                non_nan_seen = True
                new_time.append(time[j])
                new_data.append(data[j])
                if alen == 3 and append_v:
                    new_v.append(v[j])
        if non_nan_seen is False:
            logging.warning('No unflagged data in %s, returning.', tvar)
            return
        if newname is None:
            if alen == 2:
                store_data(tvar, data={'x': new_time, 'y': new_data})
            else:
                store_data(tvar, data={'x': new_time, 'y': new_data, 'v': new_v})
        else:
            if alen == 2:
                store_data(newname, data={'x': new_time, 'y': new_data})
            else:
                store_data(newname, data={'x': new_time, 'y': new_data, 'v': new_v})
            pyspedas.tplot_tools.data_quants[newname].attrs = copy.deepcopy(pyspedas.tplot_tools.data_quants[tvar].attrs)
    elif method == 'repeat' or method == 'linear' or method == 'replace':
        a = copy.deepcopy(get_data(tvar))
        time = a[0]
        data = a[1]
        # Force the data into a 2-d view, retaining the original shape so we can restore it later
        original_data_shape = data.shape
        data = data.reshape(len(time), -1)
        alen = len(a)
        if alen > 3:
            logging.info('deflag is not used for more than 2-d input')
            return
        if alen == 3:
            v = a[2]
        ntimes = time.shape[0]
        nydim = data.ndim
        if nydim == 1:
            ny = 1
        else:
            ny = data.shape[1]
        for i in range(nf):
            if np.isnan(flag[i]):  # NaN flags need special handling
                flag_is_nan = True
            else:
                flag_is_nan = False
            for k in range(ny):
                #print(data[:, k])
                if (flag_is_nan):
                    flagged_data = np.argwhere(np.isnan(data[:, k])).flatten()
                else:
                    flagged_data = np.argwhere(data[:, k] == flag[i]).flatten()
                if len(flagged_data) > 0:
                    if flag_is_nan:
                        okval = np.argwhere(np.isfinite(data[:, k])).flatten()
                    else:
                        okval = np.argwhere(data[:, k] != flag[i]).flatten()

                    if len(okval) == 0 and method in ['linear', 'repeat']:
                        logging.info('No unflagged data in %s, returning', tvar)
                        return
                    if method == 'repeat':  # flagged data repeats the previous unflagged value
                        for j in range(ntimes):
                            if (flag_is_nan):
                                if np.isnan(data[j, k]):
                                    if j == 0:
                                        data[j, k] = data[okval[0], k]
                                    else:
                                        data[j, k] = data[j-1, k]
                            else:
                                if data[j, k] == flag[i]:
                                    if j == 0:
                                        data[j, k] = data[okval[0], k]
                                    else:
                                        data[j, k] = data[j-1, k]
                        #print("After repeat: ", data[:, k])
                    elif method == 'replace':
                        if fillval is None:
                            fv = np.nan
                        else:
                            fv = fillval
                        data[flagged_data, k] = fv
                        #print("After replace: ", data[:,k])
                    else:  # method = 'linear'
                        # interpolate flagged data, using np.interp
                        dataj = data[okval, k]
                        timej = time[okval]
                        timej_flagged = time[flagged_data]
                        dataj_interpd = np.interp(timej_flagged, timej, dataj)
                        data[flagged_data, k] = dataj_interpd
                        #print("After linear: ", data[:,k])
                else:
                    logging.info("No flagged data in %s", tvar)

        if newname is None:
            if alen == 2:
                store_data(tvar, data={'x': time, 'y': data.reshape(original_data_shape)})
            else:
                store_data(tvar, data={'x': time, 'y': data.reshape(original_data_shape), 'v': v})
        else:
            if alen == 2:
                store_data(newname, data={'x': time, 'y': data.reshape(original_data_shape)})
            else:
                store_data(newname, data={'x': time, 'y': data.reshape(original_data_shape), 'v': v})
                pyspedas.tplot_tools.data_quants[newname].attrs = copy.deepcopy(pyspedas.tplot_tools.data_quants[tvar].attrs)
    else:  # any other option includes method=None, replace flags with NaN
        nf = len(flag)
        a = copy.deepcopy(pyspedas.tplot_tools.data_quants[tvar].where(pyspedas.tplot_tools.data_quants[tvar] != flag[0]))
        if nf > 1:
            for j in range(nf):
                a = copy.deepcopy(a.where(a != flag[j]))
        if newname is None:
            a.name = tvar
            pyspedas.tplot_tools.data_quants[tvar] = a
        else:
            if 'spec_bins' in a.coords:
                store_data(newname, data={'x': a.coords['time'], 'y': a.values, 'v': a.coords['spec_bins']})
                pyspedas.tplot_tools.data_quants[newname].attrs = copy.deepcopy(pyspedas.tplot_tools.data_quants[tvar].attrs)
            else:
                store_data(newname, data={'x': a.coords['time'], 'y': a.values})
                pyspedas.tplot_tools.data_quants[newname].attrs = copy.deepcopy(pyspedas.tplot_tools.data_quants[tvar].attrs)

    return



import pyspedas

from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot

from .config import CONFIG


def load_orbit(
    trange=["2013-11-5", "2013-11-6"],
    probe="15",
    prefix="",
    suffix="",
    get_support_data=False,
    varformat=None,
    varnames=[],
    downloadonly=False,
    notplot=False,
    no_update=False,
    time_clip=True,
    force_download=False,
):
    """
    Load GOES orbit data from SPDF.

    Fetches GOES orbit data from the Space Physics Data Facility (SPDF) website:
    https://spdf.gsfc.nasa.gov/pub/data/goes/goes#/orbit/YYYY/

    Parameters
    ----------
    trange : list of str
        Time range of interest ['YYYY-MM-DD', 'YYYY-MM-DD'].
        Or, to specify more or less than a day: ['YYYY-MM-DD hh:mm:ss', 'YYYY-MM-DD hh:mm:ss'].
    probe : str or int or list of str or int
        GOES spacecraft number(s), e.g., probe=15.
    prefix : str, optional
        Prefix to add to the tplot variable names.
        By default, the added prefix is 'g[probe]_orbit_'.
    suffix : str, optional
        Suffix to add to the tplot variable names. By default, no suffix is added.
    downloadonly : bool, optional
        If True, downloads the CDF files without loading them into tplot variables. Default is False.
    notplot : bool, optional
        If True, loads the data into dictionaries containing the numpy objects instead of creating tplot variables. Default is False.
    no_update : bool, optional
        If True, only loads data from the local cache. Default is False.
    time_clip : bool, optional
        If True, clips the variables to exactly the range specified in the trange keyword. Default is False.
    force_download : bool, optional
        If True, downloads the file even if a newer version exists locally. Default is False.

    Returns
    -------
    list of str
        List of tplot variables created or list of filenames downloaded.
    """
    remote_data_dir = "https://spdf.gsfc.nasa.gov/pub/data/goes/"
    out_files = []  # list of local files downloaded
    if notplot:
        tvars = {}  # dictionary of tplot structures created
    else:
        tvars = []  # list of tplot variables created

    if not isinstance(probe, list):
        probe = [probe]

    for prb in probe:
        # yearly files
        pathformat = (
            "goes"
            + str(prb)
            + "/orbit/%Y/goes"
            + str(prb)
            + "_ephemeris_ssc_%Y0101_v??.cdf"
        )

        # find the full remote path names using the trange
        remote_names = dailynames(file_format=pathformat, trange=trange)

        files = download(
            remote_file=remote_names,
            remote_path=remote_data_dir,
            local_path=CONFIG["local_data_dir"],
            no_download=no_update,
            force_download=force_download,
        )

        out_files_local = []

        if files is not None:
            for file in files:
                out_files_local.append(file)

        out_files.extend(out_files_local)

        tvars_local = []
        if not downloadonly:
            if prefix is None or prefix == "" or prefix == "probename":
                # Example of prefix: g15_orbit_
                prefix_local = "g" + str(prb) + "_" + "orbit_"
            else:
                prefix_local = prefix

            tvars_local = cdf_to_tplot(
                out_files_local,
                prefix=prefix_local,
                suffix=suffix,
                get_support_data=get_support_data,
                varformat=varformat,
                varnames=varnames,
                notplot=notplot,
            )

            # if notplot, we're merging dictionaries, otherwise we're extending lists
            if notplot:
                tvars.update(tvars_local)
            else:
                tvars.extend(tvars_local)

            # skip time_clip if notplot is specified
            if time_clip and not notplot:
                tclip(tvars_local, trange[0], trange[1], suffix="")

    if downloadonly:
        return out_files

    return tvars



import numpy as np
import pyspedas
from pyspedas import mms_load_fpi
import unittest

from pyspedas.tplot_tools import data_exists, get_data, tplot
from pyspedas.projects.mms.fpi_tools.mms_fpi_split_tensor import mms_fpi_split_tensor
from pyspedas.projects.mms.fpi_tools.mms_fpi_ang_ang import mms_fpi_ang_ang
from pyspedas.projects.mms.fpi_tools.mms_get_fpi_dist import mms_get_fpi_dist
from pyspedas.projects.mms.fpi_tools.mms_pad_fpi import mms_pad_fpi

global_display=False


class FPITestCases(unittest.TestCase):
    def test_load_default_data(self):
        data = mms_load_fpi(trange=['2015-10-16/14:00', '2015-10-16/15:00'], available=True)
        data = mms_load_fpi(trange=['2015-10-16/14:00', '2015-10-16/15:00'])
        self.assertTrue(data_exists('mms1_dis_energyspectr_omni_fast'))

    # Test loading partial moments data sets

    def test_load_partmoms_data(self):
        data = mms_load_fpi(trange=['2015-10-16/14:00', '2015-10-16/15:00'], datatype=['des-partmoms', 'dis-partmoms'])
        self.assertTrue(data_exists('mms1_des_bulkv_part_gse_fast'))
        self.assertTrue(data_exists('mms1_dis_bulkv_part_gse_fast'))

    # Regression test for non-monotonic timestamps in FPI numberdensity variables.
    # They are present in both des-moms and des-momsaux L2 CDFs, and if both sets of files
    # are passed in the same call to cdf_to_tplot, the output tplot variable can
    # get duplicate (therefore nonmonotonic) times and data values from merging the des-moms and des-momsaux
    # copies.

    def test_numberdensity_monotonic(self):
        data = mms_load_fpi(trange=['2015-10-16/14:00', '2015-10-16/15:00'])
        dens = get_data('mms1_des_numberdensity_fast')
        times = dens.times
        dt = times[1:] - times[0:-1]
        self.assertTrue(np.min(dt) > 0.0)

    def test_load_spdf_data(self):
        data = mms_load_fpi(trange=['2015-10-16/14:00', '2015-10-16/15:00'], spdf=True)
        self.assertTrue(data_exists('mms1_dis_energyspectr_omni_fast'))

    def test_load_small_brst_interval(self):
        data = mms_load_fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', datatype=['dis-moms', 'dis-dist'], time_clip=True)
        self.assertTrue(data_exists('mms1_dis_energyspectr_omni_brst'))

    def test_load_rename_bars(self):
        data = mms_load_fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', datatype='des-dist')
        data = mms_load_fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', datatype='dis-dist')
        data = mms_load_fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', datatype='des-moms')
        data = mms_load_fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', datatype='dis-moms')
        self.assertTrue(data_exists('mms1_dis_compressionloss_brst_moms'))
        self.assertTrue(data_exists('mms1_dis_errorflags_brst_moms'))
        self.assertTrue(data_exists('mms1_des_errorflags_brst_moms'))
        self.assertTrue(data_exists('mms1_des_compressionloss_brst_moms'))
        self.assertTrue(data_exists('mms1_des_errorflags_brst_dist'))
        self.assertTrue(data_exists('mms1_des_compressionloss_brst_dist'))
        self.assertTrue(data_exists('mms1_dis_errorflags_brst_dist'))
        self.assertTrue(data_exists('mms1_dis_compressionloss_brst_dist'))

    def test_center_fast_ion_data(self):
        data = mms_load_fpi(trange=['2015-10-16/14:00', '2015-10-16/15:00'])
        centered = mms_load_fpi(trange=['2015-10-16/14:00', '2015-10-16/15:00'], center_measurement=True, suffix='_centered')

        t, d = get_data('mms1_dis_bulkv_gse_fast')
        c, d = get_data('mms1_dis_bulkv_gse_fast_centered')
        self.assertTrue(np.round(c[0]-t[0], decimals=3) == 2.25)

    def test_center_fast_electron_data(self):
        data = mms_load_fpi(trange=['2015-10-16/14:00', '2015-10-16/15:00'])
        centered = mms_load_fpi(trange=['2015-10-16/14:00', '2015-10-16/15:00'], center_measurement=True, suffix='_centered')

        t, d = get_data('mms1_des_bulkv_gse_fast')
        c, d = get_data('mms1_des_bulkv_gse_fast_centered')
        self.assertTrue(np.round(c[0]-t[0], decimals=3) == 2.25)

    def test_center_brst_ion_data(self):
        data = mms_load_fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst')
        centered = mms_load_fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', center_measurement=True, suffix='_centered')

        t, d = get_data('mms1_dis_bulkv_gse_brst')
        c, d = get_data('mms1_dis_bulkv_gse_brst_centered')
        self.assertTrue(np.round(c[0]-t[0], decimals=3) == 0.075)

    def test_center_brst_electron_data(self):
        data = mms_load_fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst')
        centered = mms_load_fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', center_measurement=True, suffix='_centered')

        t, d = get_data('mms1_des_bulkv_gse_brst')
        c, d = get_data('mms1_des_bulkv_gse_brst_centered')
        self.assertTrue(np.round(c[0]-t[0], decimals=3) == 0.015)

    def test_errorflag_compression_bars(self):
        data = mms_load_fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', datatype=['des-dist', 'des-moms'])
        data = mms_load_fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', datatype=['dis-dist', 'dis-moms'])
        self.assertTrue(data_exists('mms1_des_errorflags_brst_moms_flagbars_full'))
        self.assertTrue(data_exists('mms1_des_errorflags_brst_moms_flagbars_main'))
        self.assertTrue(data_exists('mms1_des_errorflags_brst_moms_flagbars_mini'))
        self.assertTrue(data_exists('mms1_dis_errorflags_brst_moms_flagbars_full'))
        self.assertTrue(data_exists('mms1_dis_errorflags_brst_moms_flagbars_main'))
        self.assertTrue(data_exists('mms1_dis_errorflags_brst_moms_flagbars_mini'))
        self.assertTrue(data_exists('mms1_des_errorflags_brst_dist_flagbars_dist'))
        self.assertTrue(data_exists('mms1_dis_errorflags_brst_dist_flagbars_dist'))
        self.assertTrue(data_exists('mms1_des_compressionloss_brst_moms_flagbars'))
        self.assertTrue(data_exists('mms1_dis_compressionloss_brst_moms_flagbars'))
        self.assertTrue(data_exists('mms1_des_compressionloss_brst_dist_flagbars'))
        self.assertTrue(data_exists('mms1_dis_compressionloss_brst_dist_flagbars'))

    def test_errorflag_bars_bit14(self):
        # Test that the recently added bit flag for 'MMS3 Spintone DIS008 Anomaly' is
        # correctly recognized and that the error flag bars are plottable
        data = mms_load_fpi(probe='3',trange=['2019-02-09/00:32:00', '2019-02-09/00:35:00'], data_rate='brst', datatype=['des-moms', 'dis-moms'])
        data = mms_load_fpi(probe='3',trange=['2019-02-09/00:32:00', '2019-02-09/00:35:00'], data_rate='fast', datatype=['des-moms', 'dis-moms'])
        self.assertTrue(data_exists('mms3_des_errorflags_brst_moms_flagbars_full'))
        self.assertTrue(data_exists('mms3_des_errorflags_brst_moms_flagbars_main'))
        self.assertTrue(data_exists('mms3_des_errorflags_brst_moms_flagbars_mini'))
        self.assertTrue(data_exists('mms3_dis_errorflags_brst_moms_flagbars_full'))
        self.assertTrue(data_exists('mms3_dis_errorflags_brst_moms_flagbars_main'))
        self.assertTrue(data_exists('mms3_dis_errorflags_brst_moms_flagbars_mini'))
        self.assertTrue(data_exists('mms3_des_compressionloss_brst_moms_flagbars'))
        self.assertTrue(data_exists('mms3_dis_compressionloss_brst_moms_flagbars'))
        # Test that at least one sample has bit 14 (counting from 0) turned on
        d=get_data('mms3_dis_errorflags_brst_moms_flagbars_full')
        self.assertFalse(np.isnan(d.y[:,14]).all())
        tplot('mms3_des_errorflags_brst_moms_flagbars_full', display=global_display, save_png='mm3_des_flagbars_full.png')
        tplot('mms3_des_errorflags_brst_moms_flagbars_main', display=global_display, save_png='mm3_des_flagbars_main.png')
        tplot('mms3_des_errorflags_brst_moms_flagbars_mini', display=global_display, save_png='mm3_des_flagbars_mini.png')
        tplot('mms3_dis_errorflags_brst_moms_flagbars_full', display=global_display, save_png='mm3_dis_flagbars_full.png')
        tplot('mms3_dis_errorflags_brst_moms_flagbars_main', display=global_display, save_png='mm3_dis_flagbars_main.png')
        tplot('mms3_dis_errorflags_brst_moms_flagbars_mini', display=global_display, save_png='mm3_dis_flagbars_mini.png')


    def test_angle_angle(self):
        mms_fpi_ang_ang('2015-10-16/13:06:59.985', species='e',data_rate='brst', save_png='mms1_fpi_ang_ang_brst', display=False)
        mms_fpi_ang_ang('2015-10-16/13:06:30', save_jpeg='mms1_fpi_ang_ang', display=False)
        mms_fpi_ang_ang('2015-10-16/13:06:30', probe='4', save_svg='mms4_fpi_ang_ang', display=False)
        mms_fpi_ang_ang('2015-10-16/13:06:30', probe='4', save_eps='mms4_fpi_ang_ang_viridis', cmap='viridis', display=False)

    def test_pad(self):
        trange = ['2015-10-16/13:06:29', '2015-10-16/13:06:32']
        pyspedas.projects.mms.fpi(trange=trange, data_rate='brst', datatype=['dis-dist', 'des-dist', 'dis-moms'], time_clip=True)
        pyspedas.projects.mms.fgm(trange=trange, data_rate='brst')
        dists = mms_get_fpi_dist('mms1_dis_dist_brst')
        dists_e = mms_get_fpi_dist('mms1_des_dist_brst')
        pa_dist = mms_pad_fpi(dists, trange=trange, mag_data='mms1_fgm_b_gse_brst_l2_bvec')
        pa_dist = mms_pad_fpi(dists_e, trange=trange, mag_data='mms1_fgm_b_gse_brst_l2_bvec')
        pa_dist = mms_pad_fpi(dists, time='2015-10-16/13:06:30', units='eflux', mag_data='mms1_fgm_b_gse_brst_l2_bvec')
        pa_dist = mms_pad_fpi(dists,
                      subtract_bulk=True,
                      time='2015-10-16/13:06:30',
                      units='eflux',
                      mag_data='mms1_fgm_b_gse_brst_l2_bvec',
                      vel_data='mms1_dis_bulkv_gse_brst')

    def test_varname_support_data(self):
        """ Test whether support data variables are returned if explicit varname or varformat is specified """
        trange = ['2015-10-16/13:06:29', '2015-10-16/13:06:32']
        data = pyspedas.projects.mms.fpi(probe='1',trange=trange,data_rate='brst',level='l2',datatype='des-dist',varnames=['mms1_des_energy_brst'],time_clip=True)
        self.assertTrue('mms1_des_energy_brst' in data)
        data = pyspedas.projects.mms.fpi(probe='1',trange=trange,data_rate='brst',level='l2',datatype='des-dist',varformat='*des_energy_brst',time_clip=True)
        self.assertTrue('mms1_des_energy_brst' in data)

    def test_split_tensors(self):
        data = pyspedas.projects.mms.fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'],
                                data_rate='brst',
                                datatype=['dis-moms', 'des-moms'])

        split_vars = mms_fpi_split_tensor('mms1_dis_temptensor_gse_brst')

        components = ['xx', 'xy', 'xz', 'yx', 'yy', 'yz', 'zx', 'zy', 'zz']
        output = ['mms1_dis_temptensor_gse_brst_' + component for component in components]

        for v in output:
            self.assertTrue(data_exists(v))

        split_vars = mms_fpi_split_tensor('mms1_des_temptensor_gse_brst')

        components = ['xx', 'xy', 'xz', 'yx', 'yy', 'yz', 'zx', 'zy', 'zz']
        output = ['mms1_des_temptensor_gse_brst_' + component for component in components]

        for v in output:
            self.assertTrue(data_exists(v))


if __name__ == '__main__':
    unittest.main()



from .load import load
from .mfi import mfi
from .swe import swe
from .epam import epam
from .cris import cris
from .sis import sis
from .uleis import uleis
from .sepica import sepica
from .swics import swics
from .datasets import datasets


"""
Creates a new tplot variable which is the derivative of the data.

Notes
-----
Similar to deriv_data.pro in IDL SPEDAS.

"""
import logging

import numpy as np

import pyspedas


def deriv_data(names, newname=None, suffix=None, overwrite=None, edge_order=1):
    """
    Compute the derivative.

    Parameters
    ----------
    names: str/list of str
        List of pytplot names.
    newname: str/list of str, optional
        List of new names for tplot variables.
        If not given, then a suffix is applied.
    suffix: str, optional
        A suffix to apply. Default is '-avg'.
    overwrite: bool, optional
        Replace the existing tplot name.
    edge_order: int, optional
        A value passed to np.gradient that specifies how boundaries are treated

    Returns
    -------
    None.

    """

    old_names = pyspedas.tnames(names)

    if len(old_names) < 1:
        logging.error('deriv_data: No valid tplot names were provided.')
        return

    if suffix is None:
        suffix = '-der'

    if overwrite is not None:
        n_names = old_names
    elif newname is None:
        n_names = [s + suffix for s in old_names]
    else:
        n_names = newname

    if isinstance(n_names, str):
        n_names = [n_names]

    if len(n_names) != len(old_names):
        n_names = [s + suffix for s in old_names]

    for i, old in enumerate(old_names):
        data = pyspedas.get_data(old)
        data_grad = np.gradient(data.y, data.times, axis = 0, edge_order=edge_order)
        pyspedas.store_data(n_names[i], data={'x': data.times, 'y': data_grad})
        logging.info('deriv_data was applied to: ' + n_names[i])

    return n_names


import os
import unittest

import pyspedas
from pyspedas.utilities.download import download
from pyspedas.utilities.download_ftp import download_ftp

from pyspedas.projects.themis.config import CONFIG
# Use whichever remote server is configured for THEMIS, in case the Berkeley server is unavailable
themis_remote = CONFIG['remote_data_dir']


class DownloadTestCases(unittest.TestCase):
    def test_remote_path(self):
        # only specifying remote_path saves the files to the current working directory
        files = download(remote_path='https://spdf.gsfc.nasa.gov/pub/data/psp/sweap/spc/l3/l3i/2019/psp_swp_spc_l3i_20190401_v01.cdf')
        self.assertTrue(len(files) == 1)
        self.assertTrue(files[0] == os.path.join(os.getcwd(), 'psp_swp_spc_l3i_20190401_v01.cdf'))

    def test_remote_file(self):
        # only specifying remote_file saves the files to the current working directory
        files = download(remote_file='https://spdf.gsfc.nasa.gov/pub/data/psp/sweap/spc/l3/l3i/2019/psp_swp_spc_l3i_20190401_v01.cdf')
        self.assertTrue(len(files) == 1)
        self.assertTrue(files[0] == os.path.join(os.getcwd(), 'psp_swp_spc_l3i_20190401_v01.cdf'))

    def test_wildcard(self):
        # Test a wildcard pattern with several matches
        files = download(remote_path="https://themis-data.igpp.ucla.edu/tha/l1/state/2008/tha_l1_state_20080323_v??.cdf", last_version=False)
        self.assertTrue(len(files) == 4)  # v00, v01, v02, v03 should be available on this date
        self.assertTrue(files[3] == os.path.join(os.getcwd(), 'tha_l1_state_20080323_v03.cdf'))

    def test_missing_index(self):
        # Test a wildcard pattern on a nonexistent directory
        # This should warn "Remote index not found"
        files = download(remote_path=themis_remote+'l1/state/2006/tha_l1_state_20060323_v??.cdf')
        self.assertTrue(len(files) == 0)

    def test_last_version(self):
        # Test a wildcard pattern with several matches, and last_version=True, returning the final (lexicographic) value
        files = download(remote_path=themis_remote+'tha/l1/state/2008/tha_l1_state_20080323_v??.cdf',last_version=True)
        self.assertTrue(len(files) == 1)  # v00, v01, v02, v03 should be available on this date, should only return v03
        self.assertTrue(files[0] == os.path.join(os.getcwd(), 'tha_l1_state_20080323_v03.cdf'))

    def test_last_version_nomatch(self):
        # Test a wildcard pattern that doesn't match anything.
        # This should warn about no matching file found in the index.
        files = download(remote_path=themis_remote+'tha/l1/state/2008/tha_l1_state_20080332_v??.cdf',last_version=True)
        self.assertTrue(len(files) == 0)  # Nonexistent date, nothing should be found

    def test_remote_path_file(self):
        # specifying both remote_path and remote_file saves the files to the current working directory + the path specified in remote_file
        files = download(remote_path='https://spdf.gsfc.nasa.gov/pub/data/', remote_file='psp/sweap/spc/l3/l3i/2019/psp_swp_spc_l3i_20190401_v01.cdf')
        self.assertTrue(len(files) == 1)
        self.assertTrue(files[0] == os.path.join(os.getcwd(), 'psp/sweap/spc/l3/l3i/2019/psp_swp_spc_l3i_20190401_v01.cdf'))

    def test_local_file(self):
        # specifying local_file changes the local file name
        files = download(local_file='psp_data.cdf', remote_file='https://spdf.gsfc.nasa.gov/pub/data/psp/sweap/spc/l3/l3i/2019/psp_swp_spc_l3i_20190401_v01.cdf')
        self.assertTrue(len(files) == 1)
        self.assertTrue(files[0] == os.path.join(os.getcwd(), 'psp_data.cdf'))

    def test_local_path(self):
        # specifying local_path changes the local data directory
        files = download(local_path='psp_data/spc/l3', remote_file='https://spdf.gsfc.nasa.gov/pub/data/psp/sweap/spc/l3/l3i/2019/psp_swp_spc_l3i_20190401_v01.cdf')
        self.assertTrue(len(files) == 1)
        self.assertTrue(files[0] == os.path.join('psp_data/spc/l3', 'psp_swp_spc_l3i_20190401_v01.cdf'))

    def test_force_download(self):
        # specifying both remote_path and remote_file saves the files to the current working directory + the path specified in remote_file
        files = download(remote_path='https://spdf.gsfc.nasa.gov/pub/data/', remote_file='themis/tha/l1/state/2007/tha_l1_state_20070217_v01.cdf')
        self.assertTrue(len(files) == 1)
        self.assertTrue(files[0] == os.path.join(os.getcwd(), 'themis/tha/l1/state/2007/tha_l1_state_20070217_v01.cdf'))
        # Download the same file without force_download, should not re-download
        with self.assertLogs(level='INFO') as log:
            files = download(remote_path='https://spdf.gsfc.nasa.gov/pub/data/',
                             remote_file='themis/tha/l1/state/2007/tha_l1_state_20070217_v01.cdf')
            self.assertIn("File is current", log.output[0])
        # Download the same file with force_download, should re-download
        with self.assertLogs(level='INFO') as log:
            files = download(remote_path='https://spdf.gsfc.nasa.gov/pub/data/',
                             remote_file='themis/tha/l1/state/2007/tha_l1_state_20070217_v01.cdf',
                             force_download=True)
            self.assertIn("Downloading", log.output[0])

    def test_cdf_to_tplot_varformat(self):
        # Check varnamea, varformat, exclude_format, and suffixes in various combinations
        # single varformat pattern
        vars = pyspedas.projects.themis.state(probe='a',varformat='*pos*')
        vars.sort()
        self.assertEqual(vars, ['tha_pos', 'tha_pos_gse', 'tha_pos_gsm', 'tha_pos_sel', 'tha_pos_sse'])
        # single varformat pattern with "|", should include both options
        vars = pyspedas.projects.themis.state(probe='a', varformat='*pos*|*vel*')
        vars.sort()
        self.assertEqual(vars,['tha_pos', 'tha_pos_gse', 'tha_pos_gsm', 'tha_pos_sel', 'tha_pos_sse', 'tha_vel', 'tha_vel_gse', 'tha_vel_gsm', 'tha_vel_sel', 'tha_vel_sse'])
        # varformat pattern with spaces, should treat as alternate patterns
        vars = pyspedas.projects.themis.state(probe='a', varformat='*pos*  *vel*')
        vars.sort()
        self.assertEqual(vars,['tha_pos', 'tha_pos_gse', 'tha_pos_gsm', 'tha_pos_sel', 'tha_pos_sse', 'tha_vel', 'tha_vel_gse', 'tha_vel_gsm', 'tha_vel_sel', 'tha_vel_sse'])
        # varformat and exclude_format patterns with spaces
        vars = pyspedas.projects.themis.state(probe='a', varformat='*pos*  *vel*', exclude_format='*sel* *sse*')
        vars.sort()
        self.assertEqual(vars,['tha_pos', 'tha_pos_gse', 'tha_pos_gsm', 'tha_vel', 'tha_vel_gse', 'tha_vel_gsm'])
        # varformat patterns as array, should treat as alternate patterns
        vars = pyspedas.projects.themis.state(probe='a', varformat=['*pos*','*vel*'])
        vars.sort()
        self.assertEqual(vars,['tha_pos', 'tha_pos_gse', 'tha_pos_gsm', 'tha_pos_sel', 'tha_pos_sse', 'tha_vel', 'tha_vel_gse', 'tha_vel_gsm', 'tha_vel_sel', 'tha_vel_sse'])
        # varformat and exclude_format patterns as arrays
        vars = pyspedas.projects.themis.state(probe='a', varformat=['*pos*','*vel*'], exclude_format=['*sel*', '*sse*'])
        vars.sort()
        self.assertEqual(vars,['tha_pos', 'tha_pos_gse', 'tha_pos_gsm', 'tha_vel', 'tha_vel_gse', 'tha_vel_gsm'])
        # varnames with suffix specified, should return requested variables even though CDF names don't have suffix
        vars = pyspedas.projects.themis.state(probe='a',suffix='_suf', varnames=['tha_pos_suf', 'tha_pos_gse_suf'])
        vars.sort()
        self.assertEqual(vars, ['tha_pos_gse_suf', 'tha_pos_suf'])
        # varformat with suffix specified, should return values
        vars = pyspedas.projects.themis.state(probe='a',suffix='_suf', varformat=['*pos_suf', '*pos_gse_suf'])
        vars.sort()
        self.assertEqual(vars, ['tha_pos_gse_suf', 'tha_pos_suf'])
        # exclude_format with suffix specified, should remove variables with or without suffixes
        vars = pyspedas.projects.themis.state(probe='a',suffix='_suf', exclude_format=['*vel*', '*spin*','*sse*_suf', '*sel'])
        vars.sort()
        self.assertEqual(vars, ['tha_pos_gse_suf', 'tha_pos_gsm_suf', 'tha_pos_suf'])

    def test_download_ftp(self):
        from pyspedas.projects.noaa.config import CONFIG
        kp_mirror = "ftp.gfz-potsdam.de"
        remote_kp_dir = "/pub/home/obs/kp-ap/wdc/yearly/"
        local_data_dir = CONFIG['local_data_dir']
        pre = "kp"
        suf = ".wdc"
        start_year = 2019
        end_year = 2021
        # Remote names
        remote_names = [
            pre + str(year) + suf for year in range(start_year, end_year + 1)
        ]
        files=[]
        for yearstr in remote_names:
            dfile = download_ftp(
                kp_mirror,
                remote_kp_dir,
                yearstr,
                local_data_dir,
                force_download=True,
            )
            if len(dfile) > 0:
                files.append(dfile[0])
        self.assertTrue(os.path.exists(files[0]))
        self.assertTrue(os.path.exists(files[1]))
        self.assertTrue(os.path.exists(files[2]))
        # repeat download of first file to exercise 'file not modified' path
        dfile = download_ftp(
            kp_mirror,
            remote_kp_dir,
            remote_names[0],
            local_data_dir,
            force_download=False,
        )
        self.assertTrue(len(dfile) == 1)



if __name__ == "__main__":
    unittest.main()



import pandas as pd
from pyspedas.tplot_tools import time_float
from pyspedas.tplot_tools import store_data

from ..load import load


from typing import List, Optional

def att(
    trange: List[str] = ['2017-04-01', '2017-04-02'],
    level: str = 'l2',
    downloadonly: bool = False,
    notplot: bool = False,
    no_update: bool = False,
    uname: Optional[str] = None,
    passwd: Optional[str] = None,
    force_download: bool = False,
) -> List[str]:
    """
    This function loads attitude data from the Arase mission

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2017-04-01', '2017-04-02']

        level: str
            Data level; Valid options: 'l2'
            Default: 'l2'

        downloadonly: bool
            Set this flag to download the files, but not load them into
            tplot variables. Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables. Default: False

        no_update: bool
            If set, only load data from your local cache. Default: False

        uname: str
            User name. Default: None

        passwd: str
            Password. Default: None

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False

    Returns
    -------
        list of str
            List of tplot variables loaded

    Examples
    --------
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> att_vars = pyspedas.projects.erg.att(trange=['2017-04-01', '2017-04-02'])
    >>> tplot(['erg_att_sprate', 'erg_att_spphase', 'erg_att_izras', 'erg_att_izdec', 'erg_att_gxras', 'erg_att_gxdec', 'erg_att_gzras', 'erg_att_gzdec'])


    """
    file_res = 24*3600.
    pathformat = 'satellite/erg/att/txt/erg_att_'+level+'_%Y%m%d_v??.txt'

    out_files = load(pathformat=pathformat, trange=trange, file_res=file_res,
                     downloadonly=True, no_update=no_update, uname=uname, passwd=passwd, force_download=force_download)

    if downloadonly:
        return out_files

    data_flame_list = []
    for file in out_files:
        raw_read_table = pd.read_table(file)
        data_flame_list.append(
            raw_read_table[10:][raw_read_table.keys()[0]].str.split(expand=True))

    concat_frame_for_tplot = pd.concat(data_flame_list)
    time_float_array = time_float(concat_frame_for_tplot.iloc[:, 0])

    Omega_float_array = concat_frame_for_tplot.iloc[:, 1].astype(float)
    Phase_float_array = concat_frame_for_tplot.iloc[:, 9].astype(float)
    I_Alpha_float_array = concat_frame_for_tplot.iloc[:, 2].astype(float)
    I_Delta_float_array = concat_frame_for_tplot.iloc[:, 3].astype(float)
    GX_Alpha_float_array = concat_frame_for_tplot.iloc[:, 10].astype(float)
    GX_Delta_float_array = concat_frame_for_tplot.iloc[:, 11].astype(float)
    GZ_Alpha_float_array = concat_frame_for_tplot.iloc[:, 12].astype(float)
    GZ_Delta_float_array = concat_frame_for_tplot.iloc[:, 13].astype(float)

    if notplot:
        output_dictionary = {}
        output_dictionary['erg_att_sprate'] = {
            'x': time_float_array, 'y': Omega_float_array}
        output_dictionary['erg_att_spphase'] = {
            'x': time_float_array, 'y': Phase_float_array}
        output_dictionary['erg_att_izras'] = {
            'x': time_float_array, 'y': I_Alpha_float_array}
        output_dictionary['erg_att_izdec'] = {
            'x': time_float_array, 'y': I_Delta_float_array}
        output_dictionary['erg_att_gxras'] = {
            'x': time_float_array, 'y': GX_Alpha_float_array}
        output_dictionary['erg_att_gxdec'] = {
            'x': time_float_array, 'y': GX_Delta_float_array}
        output_dictionary['erg_att_gzras'] = {
            'x': time_float_array, 'y': GZ_Alpha_float_array}
        output_dictionary['erg_att_gzdec'] = {
            'x': time_float_array, 'y': GZ_Delta_float_array}
        return output_dictionary
    else:
        store_data('erg_att_sprate', data={
            'x': time_float_array, 'y': Omega_float_array})
        store_data('erg_att_spphase', data={
            'x': time_float_array, 'y': Phase_float_array})
        store_data('erg_att_izras', data={
            'x': time_float_array, 'y': I_Alpha_float_array})
        store_data('erg_att_izdec', data={
            'x': time_float_array, 'y': I_Delta_float_array})
        store_data('erg_att_gxras', data={
            'x': time_float_array, 'y': GX_Alpha_float_array})
        store_data('erg_att_gxdec', data={
            'x': time_float_array, 'y': GX_Delta_float_array})
        store_data('erg_att_gzras', data={
            'x': time_float_array, 'y': GZ_Alpha_float_array})
        store_data('erg_att_gzdec', data={
            'x': time_float_array, 'y': GZ_Delta_float_array})

    return ['erg_att_sprate','erg_att_spphase','erg_att_izras','erg_att_gxras','erg_att_gxdec','erg_att_gzras','erg_att_gzdec']



import cdflib
from pyspedas.tplot_tools import get_data
import logging

def get_gatt_ror(downloadonly, loaded_data):
    """Get global attributes from a downloaded CDF file or loaded tplot variable, in order to print rules of the road"""
    try:
        if isinstance(loaded_data, list):
            if downloadonly:
                cdf_file = cdflib.CDF(loaded_data[-1])
                gatt = cdf_file.globalattsget()
            else:
                md = get_data(loaded_data[-1], metadata=True)
                if type(md) is not dict:
                    # If the last variable doesn't work, try the first variable
                    md = get_data(loaded_data[0], metadata=True)
                gatt = md['CDF']['GATT']
        elif isinstance(loaded_data, dict):
            gatt = loaded_data[list(loaded_data.keys())[-1]]['CDF']['GATT']
        return gatt
    except:
        logging.warning('Retrieving PI info and rules of the road failed')
        # Return a dictionary containing all the fields the various load routines are looking for
        return {'PI_name':'Error_retrieving',
                "PI_NAME":"Error_retrieving",
                'PI_affiliation':'Error_retrieving',
                "PI_AFFILIATION":"Error_retrieving",
                'LINK':'Error_retrieving',
                "TEXT":["Error_retrieving"],
                "LINK_TEXT":"Error_retriving",
                "HTTP_LINK":"Error_retrieving",
                "Rules_of_use":"Error_retrieving",
                "Logical_source_description":"Error_retrieving",
                "LOGICAL_SOURCE_DESCRIPTION":"Error_retrieving",
                "Station_code":"Error_retrieving"
                }
